import functools
import logging
from typing import Any, Union

from pydantic import ConfigDict, PrivateAttr

from magnus import defaults
from magnus.experiment_tracker import BaseExperimentTracker

logger = logging.getLogger(defaults.NAME)


class MLFlowExperimentTracker(BaseExperimentTracker):
    """
    A MLFlow experiment tracker.

    TODO: Need to set up credentials from secrets
    """

    service_name: str = "mlflow"

    server_url: str
    autolog: bool = False

    _default_experiment_name: str = PrivateAttr(default="Default")
    _active_run_id: str = PrivateAttr(default="")
    _client: Any = PrivateAttr(default=None)

    model_config = ConfigDict(extra="forbid")

    def model_post_init(self, __context: Any) -> None:
        try:
            import mlflow
        except ImportError:
            raise Exception("You need to install mlflow to use MLFlowExperimentTracker.")

        self._client = mlflow

        self._client.set_tracking_uri(self.server_url)

        if self.autolog:
            self._client.autolog(log_models=False)

    @functools.cached_property
    def experiment_id(self):
        experiment_name = self._default_experiment_name

        # If a tag is provided, we should create that as our experiment
        if self._context.tag:
            experiment_name = self._context.tag

        experiment = self._client.get_experiment_by_name(experiment_name)
        if not experiment:
            # Create the experiment and get it.
            experiment = self._client.create_experiment(experiment_name)
            experiment = self._client.get_experiment(experiment)

        return experiment.experiment_id

    @functools.cached_property
    def run_name(self):
        return self._context.run_id

    @property
    def client_context(self):
        if self._active_run_id:
            return self._client.start_run(
                run_id=self._active_run_id, experiment_id=self.experiment_id, run_name=self.run_name
            )

        active_run = self._client.start_run(run_name=self.run_name, experiment_id=self.experiment_id)
        self._active_run_id = active_run.info.run_id
        return active_run

    def log_metric(self, key: str, value: Union[int, float], step: int = 0):
        """
        Sets the metric in the experiment tracking.

        Args:
            key (str): The key against you want to store the value
            value (Any): The value of the metric
        """
        if not isinstance(value, float) or isinstance(value, int):
            msg = f"Only float/int values are accepted as metrics. Setting the metric {key} as parameter {key}_{step}"
            logger.warning(msg)
            self.log_parameter(key=key, value=value, step=step)
            return

        with self.client_context as _:
            self._client.log_metric(key, float(value), step=step or None)

    def log_parameter(self, key: str, value: Any, step: int = 0):
        with self.client_context as _:
            self._client.log_param(key + f"_{str(step)}", value)



# ruff: noqa

# TODO: Might need to add Rich to pyinstaller part
import logging
from logging.config import dictConfig

from magnus import defaults

dictConfig(defaults.LOGGING_CONFIG)
logger = logging.getLogger(defaults.LOGGER_NAME)

from magnus.interaction import (
    end_interactive_session,
    get_experiment_tracker_context,
    get_from_catalog,
    get_object,
    get_parameter,
    get_run_id,
    get_run_log,
    get_secret,
    put_in_catalog,
    put_object,
    start_interactive_session,
    set_parameter,
    track_this,
)  # noqa
from magnus.sdk import Stub, Pipeline, Task, Parallel, Map, Catalog, Success, Fail  # noqa


# TODO: Think of model registry as a central place to store models.
# TODO: Implement Sagemaker pipelines as a executor.


# TODO: Think of way of generating dag hash without executor configuration



import json
import logging
import random
import shlex
import string
from abc import ABC, abstractmethod
from collections import OrderedDict
from typing import Any, Dict, List, Optional, Union, cast

from pydantic import BaseModel, ConfigDict, Field, computed_field, field_serializer, field_validator
from pydantic.functional_serializers import PlainSerializer
from ruamel.yaml import YAML
from typing_extensions import Annotated

from magnus import defaults, exceptions, integration, parameters, utils
from magnus.defaults import TypeMapVariable
from magnus.extensions.executor import GenericExecutor
from magnus.extensions.nodes import DagNode, MapNode, ParallelNode
from magnus.graph import Graph, create_node, search_node_by_internal_name
from magnus.integration import BaseIntegration
from magnus.nodes import BaseNode

logger = logging.getLogger(defaults.NAME)

# TODO: Leave the run log in consistent state.

"""
executor:
  type: argo
  config:
    image: # apply to template
    max_workflow_duration: # Apply to spec
    nodeSelector: #Apply to spec
    parallelism: #apply to spec
    resources: # convert to podSpecPath
      limits:
      requests:
    retryStrategy:
    max_step_duration: # apply to templateDefaults
    step_timeout: # apply to templateDefaults
    tolerations: # apply to spec
    imagePullPolicy: # apply to template

    overrides:
      override:
        tolerations: # template
        image: # container
        max_step_duration: # template
        step_timeout: #template
        nodeSelector: #template
        parallelism: # this need to applied for map
        resources: # container
        imagePullPolicy: #container
        retryStrategy: # template
"""


class SecretEnvVar(BaseModel):
    """
    Renders:
    env:
      - name: MYSECRETPASSWORD
        valueFrom:
          secretKeyRef:
            name: my-secret
            key: mypassword
    """

    environment_variable: str = Field(serialization_alias="name")
    secret_name: str = Field(exclude=True)
    secret_key: str = Field(exclude=True)

    @computed_field  # type: ignore
    @property
    def valueFrom(self) -> Dict[str, Dict[str, str]]:
        return {
            "secretKeyRef": {
                "name": self.secret_name,
                "key": self.secret_key,
            }
        }


class EnvVar(BaseModel):
    """
    Renders:
    parameters: # in arguments
      - name: x
        value: 3 # This is optional for workflow parameters

    """

    name: str
    value: Union[str, int, float] = Field(default="")


class Parameter(BaseModel):
    name: str
    value: Optional[str] = None

    @field_serializer("name")
    def serialize_name(self, name: str) -> str:
        return f"{str(name)}"

    @field_serializer("value")
    def serialize_value(self, value: str) -> str:
        return f"{value}"


class OutputParameter(Parameter):
    """
    Renders:
    - name: step-name
      valueFrom:
        path: /tmp/output.txt
    """

    path: str = Field(default="/tmp/output.txt", exclude=True)

    @computed_field  # type: ignore
    @property
    def valueFrom(self) -> Dict[str, str]:
        return {"path": self.path}


class Argument(BaseModel):
    """
    Templates are called with arguments, which become inputs for the template
    Renders:
    arguments:
      parameters:
        - name: The name of the parameter
          value: The value of the parameter
    """

    name: str
    value: str

    @field_serializer("name")
    def serialize_name(self, name: str) -> str:
        return f"{str(name)}"

    @field_serializer("value")
    def serialize_value(self, value: str) -> str:
        return f"{value}"


class Request(BaseModel):
    """
    The default requests
    """

    memory: str = "1Gi"
    cpu: str = "250m"


VendorGPU = Annotated[
    Optional[int],
    PlainSerializer(lambda x: str(x), return_type=str, when_used="unless-none"),
]


class Limit(Request):
    """
    The default limits
    """

    gpu: VendorGPU = Field(default=None, serialization_alias="nvidia.com/gpu")


class Resources(BaseModel):
    limits: Limit = Field(default=Limit(), serialization_alias="limits")
    requests: Request = Field(default=Request(), serialization_alias="requests")


class BackOff(BaseModel):
    duration_in_seconds: int = Field(default=2 * 60, serialization_alias="duration")
    factor: float = Field(default=2, serialization_alias="factor")
    max_duration: int = Field(default=60 * 60, serialization_alias="maxDuration")

    @field_serializer("duration_in_seconds")
    def cast_duration_as_str(self, duration_in_seconds: int, _info) -> str:
        return str(duration_in_seconds)

    @field_serializer("max_duration")
    def cast_mas_duration_as_str(self, max_duration: int, _info) -> str:
        return str(max_duration)


class Retry(BaseModel):
    limit: int = 0
    retry_policy: str = Field(default="Always", serialization_alias="retryPolicy")
    back_off: BackOff = Field(default=BackOff(), serialization_alias="backoff")

    @field_serializer("limit")
    def cast_limit_as_str(self, limit: int, _info) -> str:
        return str(limit)


class Toleration(BaseModel):
    effect: str
    key: str
    operator: str
    value: str


class TemplateDefaults(BaseModel):
    max_step_duration: int = Field(
        default=60 * 60 * 2,
        serialization_alias="activeDeadlineSeconds",
        gt=0,
        description="Max run time of a step",
    )

    @computed_field  # type: ignore
    @property
    def timeout(self) -> str:
        return f"{self.max_step_duration + 60*60}s"


ShlexCommand = Annotated[str, PlainSerializer(lambda x: shlex.split(x), return_type=List[str])]


class Container(BaseModel):
    image: str
    command: ShlexCommand
    volume_mounts: Optional[List["ContainerVolume"]] = Field(default=None, serialization_alias="volumeMounts")
    image_pull_policy: str = Field(default="", serialization_alias="imagePullPolicy")
    resources: Optional[Resources] = Field(default=None, serialization_alias="resources")

    env_vars: List[EnvVar] = Field(default_factory=list, exclude=True)
    secrets_from_k8s: List[SecretEnvVar] = Field(default_factory=list, exclude=True)

    @computed_field  # type: ignore
    @property
    def env(self) -> Optional[List[Union[EnvVar, SecretEnvVar]]]:
        if not self.env_vars and not self.secrets_from_k8s:
            return None

        return self.env_vars + self.secrets_from_k8s


class DagTaskTemplate(BaseModel):
    """
    dag:
        tasks:
          name: A
            template: nested-diamond
            arguments:
                parameters: [{name: message, value: A}]
    """

    name: str
    template: str
    depends: List[str] = []
    arguments: Optional[List[Argument]] = Field(default=None)
    with_param: Optional[str] = Field(default=None, serialization_alias="withParam")

    @field_serializer("depends")
    def transform_depends_as_str(self, depends: List[str]) -> str:
        return " || ".join(depends)

    @field_serializer("arguments", when_used="unless-none")
    def empty_arguments_to_none(self, arguments: List[Argument]) -> Dict[str, List[Argument]]:
        return {"parameters": arguments}


class ContainerTemplate(BaseModel):
    # These templates are used for actual execution nodes.
    name: str
    active_deadline_seconds: Optional[int] = Field(default=None, serialization_alias="activeDeadlineSeconds", gt=0)
    node_selector: Optional[Dict[str, str]] = Field(default=None, serialization_alias="nodeSelector")
    retry_strategy: Optional[Retry] = Field(default=None, serialization_alias="retryStrategy")
    tolerations: Optional[List[Toleration]] = Field(default=None, serialization_alias="tolerations")

    container: Container

    outputs: Optional[List[OutputParameter]] = Field(default=None, serialization_alias="outputs")
    inputs: Optional[List[Parameter]] = Field(default=None, serialization_alias="inputs")

    def __hash__(self):
        return hash(self.name)

    @field_serializer("outputs", when_used="unless-none")
    def reshape_outputs(self, outputs: List[OutputParameter]) -> Dict[str, List[OutputParameter]]:
        return {"parameters": outputs}

    @field_serializer("inputs", when_used="unless-none")
    def reshape_inputs(self, inputs: List[Parameter]) -> Dict[str, List[Parameter]]:
        return {"parameters": inputs}


class DagTemplate(BaseModel):
    # These are used for parallel, map nodes dag definition
    name: str = "magnus-dag"
    tasks: List[DagTaskTemplate] = Field(default=[], exclude=True)
    inputs: Optional[List[Parameter]] = Field(default=None, serialization_alias="inputs")
    parallelism: Optional[int] = None
    fail_fast: bool = Field(default=True, serialization_alias="failFast")

    @field_validator("parallelism")
    @classmethod
    def validate_parallelism(cls, parallelism: Optional[int]) -> Optional[int]:
        if parallelism is not None and parallelism <= 0:
            raise ValueError("Parallelism must be a positive integer greater than 0")
        return parallelism

    @computed_field  # type: ignore
    @property
    def dag(self) -> Dict[str, List[DagTaskTemplate]]:
        return {"tasks": self.tasks}

    @field_serializer("inputs", when_used="unless-none")
    def reshape_inputs(self, inputs: List[Parameter], _info) -> Dict[str, List[Parameter]]:
        return {"parameters": inputs}


class Volume(BaseModel):
    """
    spec config requires, name and persistentVolumeClaim
    step requires name and mountPath
    """

    name: str
    claim: str = Field(exclude=True)
    mount_path: str = Field(serialization_alias="mountPath", exclude=True)

    @computed_field  # type: ignore
    @property
    def persistentVolumeClaim(self) -> Dict[str, str]:
        return {"claimName": self.claim}


class ContainerVolume(BaseModel):
    name: str
    mount_path: str = Field(serialization_alias="mountPath")


class UserVolumeMounts(BaseModel):
    """
    The volume specification as user defines it.
    """

    name: str  # This is the name of the PVC on K8s
    mount_path: str  # This is mount path on the container


class NodeRenderer(ABC):
    allowed_node_types: List[str] = []

    def __init__(self, executor: "ArgoExecutor", node: BaseNode) -> None:
        self.executor = executor
        self.node = node

    @abstractmethod
    def render(self, list_of_iter_values: Optional[List] = None):
        pass


class ExecutionNode(NodeRenderer):
    allowed_node_types = ["task", "stub", "success", "fail"]

    def render(self, list_of_iter_values: Optional[List] = None):
        """
        Compose the map variable and create the execution command.
        Create an input to the command.
        create_container_template : creates an argument for the list of iter values
        """
        map_variable = self.executor.compose_map_variable(list_of_iter_values)
        command = utils.get_node_execution_command(
            self.node,
            over_write_run_id=self.executor._run_id_placeholder,
            map_variable=map_variable,
        )

        inputs = []
        if list_of_iter_values:
            for val in list_of_iter_values:
                inputs.append(Parameter(name=val))

        # Create the container template
        container_template = self.executor.create_container_template(
            working_on=self.node,
            command=command,
            inputs=inputs,
        )

        self.executor._container_templates.append(container_template)


class DagNodeRenderer(NodeRenderer):
    allowed_node_types = ["dag"]

    def render(self, list_of_iter_values: Optional[List] = None):
        self.node = cast(DagNode, self.node)
        task_template_arguments = []
        dag_inputs = []
        if list_of_iter_values:
            for value in list_of_iter_values:
                task_template_arguments.append(Argument(name=value, value="{{inputs.parameters." + value + "}}"))
                dag_inputs.append(Parameter(name=value))

        clean_name = self.executor.get_clean_name(self.node)
        fan_out_template = self.executor._create_fan_out_template(
            composite_node=self.node, list_of_iter_values=list_of_iter_values
        )
        fan_out_template.arguments = task_template_arguments if task_template_arguments else None

        fan_in_template = self.executor._create_fan_in_template(
            composite_node=self.node, list_of_iter_values=list_of_iter_values
        )
        fan_in_template.arguments = task_template_arguments if task_template_arguments else None

        self.executor._gather_task_templates_of_dag(
            self.node.branch,
            dag_name=f"{clean_name}-branch",
            list_of_iter_values=list_of_iter_values,
        )

        branch_template = DagTaskTemplate(
            name=f"{clean_name}-branch",
            template=f"{clean_name}-branch",
            arguments=task_template_arguments if task_template_arguments else None,
        )
        branch_template.depends.append(f"{clean_name}-fan-out.Succeeded")
        fan_in_template.depends.append(f"{clean_name}-branch.Succeeded")
        fan_in_template.depends.append(f"{clean_name}-branch.Failed")

        self.executor._dag_templates.append(
            DagTemplate(
                tasks=[fan_out_template, branch_template, fan_in_template],
                name=clean_name,
                inputs=dag_inputs if dag_inputs else None,
            )
        )


class ParallelNodeRender(NodeRenderer):
    allowed_node_types = ["parallel"]

    def render(self, list_of_iter_values: Optional[List] = None):
        self.node = cast(ParallelNode, self.node)
        task_template_arguments = []
        dag_inputs = []
        if list_of_iter_values:
            for value in list_of_iter_values:
                task_template_arguments.append(Argument(name=value, value="{{inputs.parameters." + value + "}}"))
                dag_inputs.append(Parameter(name=value))

        clean_name = self.executor.get_clean_name(self.node)
        fan_out_template = self.executor._create_fan_out_template(
            composite_node=self.node, list_of_iter_values=list_of_iter_values
        )
        fan_out_template.arguments = task_template_arguments if task_template_arguments else None

        fan_in_template = self.executor._create_fan_in_template(
            composite_node=self.node, list_of_iter_values=list_of_iter_values
        )
        fan_in_template.arguments = task_template_arguments if task_template_arguments else None

        branch_templates = []
        for name, branch in self.node.branches.items():
            branch_name = self.executor.sanitize_name(name)
            self.executor._gather_task_templates_of_dag(
                branch,
                dag_name=f"{clean_name}-{branch_name}",
                list_of_iter_values=list_of_iter_values,
            )
            task_template = DagTaskTemplate(
                name=f"{clean_name}-{branch_name}",
                template=f"{clean_name}-{branch_name}",
                arguments=task_template_arguments if task_template_arguments else None,
            )
            task_template.depends.append(f"{clean_name}-fan-out.Succeeded")
            fan_in_template.depends.append(f"{task_template.name}.Succeeded")
            fan_in_template.depends.append(f"{task_template.name}.Failed")
            branch_templates.append(task_template)

        executor_config = self.executor._resolve_executor_config(self.node)

        self.executor._dag_templates.append(
            DagTemplate(
                tasks=[fan_out_template] + branch_templates + [fan_in_template],
                name=clean_name,
                inputs=dag_inputs if dag_inputs else None,
                parallelism=executor_config.get("parallelism", None),
            )
        )


class MapNodeRender(NodeRenderer):
    allowed_node_types = ["map"]

    def render(self, list_of_iter_values: Optional[List] = None):
        self.node = cast(MapNode, self.node)
        task_template_arguments = []
        dag_inputs = []
        if list_of_iter_values:
            for value in list_of_iter_values:
                task_template_arguments.append(Argument(name=value, value="{{inputs.parameters." + value + "}}"))
                dag_inputs.append(Parameter(name=value))

        clean_name = self.executor.get_clean_name(self.node)
        fan_out_template = self.executor._create_fan_out_template(
            composite_node=self.node, list_of_iter_values=list_of_iter_values
        )
        fan_out_template.arguments = task_template_arguments if task_template_arguments else None

        fan_in_template = self.executor._create_fan_in_template(
            composite_node=self.node, list_of_iter_values=list_of_iter_values
        )
        fan_in_template.arguments = task_template_arguments if task_template_arguments else None

        if not list_of_iter_values:
            list_of_iter_values = []

        list_of_iter_values.append(self.node.iterate_as)

        self.executor._gather_task_templates_of_dag(
            self.node.branch,
            dag_name=f"{clean_name}-map",
            list_of_iter_values=list_of_iter_values,
        )

        task_template = DagTaskTemplate(
            name=f"{clean_name}-map",
            template=f"{clean_name}-map",
            arguments=task_template_arguments if task_template_arguments else None,
        )
        task_template.with_param = "{{tasks." + f"{clean_name}-fan-out" + ".outputs.parameters." + "iterate-on" + "}}"

        argument = Argument(name=self.node.iterate_as, value="{{item}}")
        if task_template.arguments is None:
            task_template.arguments = []
        task_template.arguments.append(argument)

        task_template.depends.append(f"{clean_name}-fan-out.Succeeded")
        fan_in_template.depends.append(f"{clean_name}-map.Succeeded")
        fan_in_template.depends.append(f"{clean_name}-map.Failed")

        executor_config = self.executor._resolve_executor_config(self.node)

        self.executor._dag_templates.append(
            DagTemplate(
                tasks=[fan_out_template, task_template, fan_in_template],
                name=clean_name,
                inputs=dag_inputs if dag_inputs else None,
                parallelism=executor_config.get("parallelism", None),
                fail_fast=executor_config.get("fail_fast", True),
            )
        )


def get_renderer(node):
    renderers = NodeRenderer.__subclasses__()

    for renderer in renderers:
        if node.node_type in renderer.allowed_node_types:
            return renderer
    raise Exception("This node type is not render-able")


class MetaData(BaseModel):
    generate_name: str = Field(default="magnus-dag-", serialization_alias="generateName")
    annotations: Optional[Dict[str, str]] = Field(default_factory=dict)
    labels: Optional[Dict[str, str]] = Field(default_factory=dict)
    namespace: Optional[str] = Field(default=None)


class Spec(BaseModel):
    active_deadline_seconds: int = Field(serialization_alias="activeDeadlineSeconds")
    entrypoint: str = Field(default="magnus-dag")
    node_selector: Optional[Dict[str, str]] = Field(default_factory=dict, serialization_alias="nodeSelector")
    tolerations: Optional[List[Toleration]] = Field(default=None, serialization_alias="tolerations")
    parallelism: Optional[int] = Field(default=None, serialization_alias="parallelism")
    # TODO: This has to be user driven
    pod_gc: Dict[str, str] = Field(default={"strategy": "OnPodCompletion"}, serialization_alias="podGC")

    retry_strategy: Retry = Field(default=Retry(), serialization_alias="retryStrategy")
    service_account_name: Optional[str] = Field(default=None, serialization_alias="serviceAccountName")

    templates: List[Union[DagTemplate, ContainerTemplate]] = Field(default_factory=list)
    template_defaults: Optional[TemplateDefaults] = Field(default=None, serialization_alias="templateDefaults")

    arguments: Optional[List[EnvVar]] = Field(default_factory=list)
    persistent_volumes: List[UserVolumeMounts] = Field(default_factory=list, exclude=True)

    @field_validator("parallelism")
    @classmethod
    def validate_parallelism(cls, parallelism: Optional[int]) -> Optional[int]:
        if parallelism is not None and parallelism <= 0:
            raise ValueError("Parallelism must be a positive integer greater than 0")
        return parallelism

    @computed_field  # type: ignore
    @property
    def volumes(self) -> List[Volume]:
        volumes: List[Volume] = []
        claim_names = {}
        for i, user_volume in enumerate(self.persistent_volumes):
            if user_volume.name in claim_names:
                raise Exception(f"Duplicate claim name {user_volume.name}")
            claim_names[user_volume.name] = user_volume.name

            volume = Volume(name=f"executor-{i}", claim=user_volume.name, mount_path=user_volume.mount_path)
            volumes.append(volume)
        return volumes

    @field_serializer("arguments", when_used="unless-none")
    def reshape_arguments(self, arguments: List[EnvVar], _info) -> Dict[str, List[EnvVar]]:
        return {"parameters": arguments}


class Workflow(BaseModel):
    api_version: str = Field(
        default="argoproj.io/v1alpha1",
        serialization_alias="apiVersion",
    )
    kind: str = "Workflow"
    metadata: MetaData = Field(default=MetaData())
    spec: Spec


class Override(BaseModel):
    model_config = ConfigDict(extra="ignore")

    image: str
    tolerations: Optional[List[Toleration]] = Field(default=None)

    max_step_duration_in_seconds: int = Field(
        default=2 * 60 * 60,  # 2 hours
        gt=0,
    )

    node_selector: Optional[Dict[str, str]] = Field(
        default=None,
        serialization_alias="nodeSelector",
    )

    parallelism: Optional[int] = Field(
        default=None,
        serialization_alias="parallelism",
    )

    resources: Resources = Field(
        default=Resources(),
        serialization_alias="resources",
    )

    image_pull_policy: str = Field(default="")

    retry_strategy: Retry = Field(
        default=Retry(),
        serialization_alias="retryStrategy",
        description="Common across all templates",
    )

    @field_validator("parallelism")
    @classmethod
    def validate_parallelism(cls, parallelism: Optional[int]) -> Optional[int]:
        if parallelism is not None and parallelism <= 0:
            raise ValueError("Parallelism must be a positive integer greater than 0")
        return parallelism


class ArgoExecutor(GenericExecutor):
    service_name: str = "argo"

    model_config = ConfigDict(extra="forbid")

    image: str
    expose_parameters_as_inputs: bool = True
    secrets_from_k8s: List[SecretEnvVar] = Field(default_factory=list)
    output_file: str = "argo-pipeline.yaml"

    # Metadata related fields
    name: str = Field(default="magnus-dag-", description="Used as an identifier for the workflow")
    annotations: Dict[str, str] = Field(default_factory=dict)
    labels: Dict[str, str] = Field(default_factory=dict)

    max_workflow_duration_in_seconds: int = Field(
        2 * 24 * 60 * 60,  # 2 days
        serialization_alias="activeDeadlineSeconds",
        gt=0,
    )
    node_selector: Optional[Dict[str, str]] = Field(
        default=None,
        serialization_alias="nodeSelector",
    )
    parallelism: Optional[int] = Field(
        default=None,
        serialization_alias="parallelism",
    )
    resources: Resources = Field(
        default=Resources(),
        serialization_alias="resources",
        exclude=True,
    )
    retry_strategy: Retry = Field(
        default=Retry(),
        serialization_alias="retryStrategy",
        description="Common across all templates",
    )
    max_step_duration_in_seconds: int = Field(
        default=2 * 60 * 60,  # 2 hours
        gt=0,
    )
    tolerations: Optional[List[Toleration]] = Field(default=None)
    image_pull_policy: str = Field(default="")
    service_account_name: Optional[str] = None
    persistent_volumes: List[UserVolumeMounts] = Field(default_factory=list)

    _run_id_placeholder: str = "{{workflow.parameters.run_id}}"
    _container_templates: List[ContainerTemplate] = []
    _dag_templates: List[DagTemplate] = []
    _clean_names: Dict[str, str] = {}
    _container_volumes: List[ContainerVolume] = []

    @field_validator("parallelism")
    @classmethod
    def validate_parallelism(cls, parallelism: Optional[int]) -> Optional[int]:
        if parallelism is not None and parallelism <= 0:
            raise ValueError("Parallelism must be a positive integer greater than 0")
        return parallelism

    @computed_field  # type: ignore
    @property
    def step_timeout(self) -> int:
        """
        Maximum time the step can take to complete, including the pending state.
        """
        return self.max_step_duration_in_seconds + 2 * 60 * 60  # 2 hours + max_step_duration_in_seconds

    @property
    def metadata(self) -> MetaData:
        return MetaData(
            generate_name=self.name,
            annotations=self.annotations,
            labels=self.labels,
        )

    @property
    def spec(self) -> Spec:
        return Spec(
            active_deadline_seconds=self.max_workflow_duration_in_seconds,
            node_selector=self.node_selector,
            tolerations=self.tolerations,
            parallelism=self.parallelism,
            retry_strategy=self.retry_strategy,
            service_account_name=self.service_account_name,
            persistent_volumes=self.persistent_volumes,
            template_defaults=TemplateDefaults(max_step_duration=self.max_step_duration_in_seconds),
        )

    def prepare_for_graph_execution(self):
        """
        This method should be called prior to calling execute_graph.
        Perform any steps required before doing the graph execution.

        The most common implementation is to prepare a run log for the run if the run uses local interactive compute.

        But in cases of actual rendering the job specs (eg: AWS step functions, K8's) we check if the services are OK.
        We do not set up a run log as its not relevant.
        """

        integration.validate(self, self._context.run_log_store)
        integration.configure_for_traversal(self, self._context.run_log_store)

        integration.validate(self, self._context.catalog_handler)
        integration.configure_for_traversal(self, self._context.catalog_handler)

        integration.validate(self, self._context.secrets_handler)
        integration.configure_for_traversal(self, self._context.secrets_handler)

        integration.validate(self, self._context.experiment_tracker)
        integration.configure_for_traversal(self, self._context.experiment_tracker)

    def prepare_for_node_execution(self):
        """
        Perform any modifications to the services prior to execution of the node.

        Args:
            node (Node): [description]
            map_variable (dict, optional): [description]. Defaults to None.
        """

        super().prepare_for_node_execution()
        self._set_up_run_log(exists_ok=True)

    def execute_node(self, node: BaseNode, map_variable: TypeMapVariable = None, **kwargs):
        step_log = self._context.run_log_store.create_step_log(node.name, node._get_step_log_name(map_variable))

        self.add_code_identities(node=node, step_log=step_log)

        step_log.step_type = node.node_type
        step_log.status = defaults.PROCESSING
        self._context.run_log_store.add_step_log(step_log, self._context.run_id)

        super()._execute_node(node, map_variable=map_variable, **kwargs)

        # Implicit fail
        if self._context.dag:
            # functions and notebooks do not have dags
            _, current_branch = search_node_by_internal_name(dag=self._context.dag, internal_name=node.internal_name)
            _, next_node_name = self._get_status_and_next_node_name(node, current_branch, map_variable=map_variable)
            if next_node_name:
                # Terminal nodes do not have next node name
                next_node = current_branch.get_node_by_name(next_node_name)

                if next_node.node_type == defaults.FAIL:
                    self.execute_node(next_node, map_variable=map_variable)

        step_log = self._context.run_log_store.get_step_log(node._get_step_log_name(map_variable), self._context.run_id)
        if step_log.status == defaults.FAIL:
            raise Exception(f"Step {node.name} failed")

    def fan_out(self, node: BaseNode, map_variable: TypeMapVariable = None):
        super().fan_out(node, map_variable)

        # If its a map node, write the list values to "/tmp/output.txt"
        if node.node_type == "map":
            node = cast(MapNode, node)
            iterate_on = self._context.run_log_store.get_parameters(self._context.run_id)[node.iterate_on]

            with open("/tmp/output.txt", mode="w", encoding="utf-8") as myfile:
                json.dump(iterate_on, myfile, indent=4)

    def _get_parameters(self) -> Dict[str, Any]:
        params = {}
        if self._context.parameters_file:
            # Parameters from the parameters file if defined
            params.update(utils.load_yaml(self._context.parameters_file))
        # parameters from environment variables supersede file based
        params.update(parameters.get_user_set_parameters())

        return params

    def sanitize_name(self, name):
        return name.replace(" ", "-").replace(".", "-").replace("_", "-")

    def get_clean_name(self, node: BaseNode):
        # Cache names for the node
        if node.internal_name not in self._clean_names:
            sanitized = self.sanitize_name(node.name)
            tag = "".join(random.choices(string.ascii_lowercase + string.digits, k=6))
            self._clean_names[node.internal_name] = f"{sanitized}-{node.node_type}-{tag}"

        return self._clean_names[node.internal_name]

    def compose_map_variable(self, list_of_iter_values: Optional[List] = None) -> TypeMapVariable:
        map_variable = OrderedDict()

        # If we are inside a map node, compose a map_variable
        # The values of "iterate_as" are sent over as inputs to the container template
        if list_of_iter_values:
            for var in list_of_iter_values:
                map_variable[var] = "{{inputs.parameters." + str(var) + "}}"

        return map_variable  # type: ignore

    def create_container_template(
        self,
        working_on: BaseNode,
        command: str,
        inputs: Optional[List] = None,
        outputs: Optional[List] = None,
        overwrite_name: str = "",
    ):
        effective_node_config = self._resolve_executor_config(working_on)

        override: Override = Override(**effective_node_config)

        container = Container(
            command=command,
            image=override.image,
            volume_mounts=self._container_volumes,
            image_pull_policy=override.image_pull_policy,
            resources=override.resources,
            secrets_from_k8s=self.secrets_from_k8s,
        )

        if working_on.name == self._context.dag.start_at and self.expose_parameters_as_inputs:
            for key, value in self._get_parameters().items():
                # Get the value from work flow parameters for dynamic behavior
                if isinstance(value, int) or isinstance(value, float) or isinstance(value, str):
                    env_var = EnvVar(
                        name=defaults.PARAMETER_PREFIX + key,
                        value="{{workflow.parameters." + key + "}}",
                    )
                    container.env_vars.append(env_var)

        clean_name = self.get_clean_name(working_on)
        if overwrite_name:
            clean_name = overwrite_name

        container_template = ContainerTemplate(
            name=clean_name,
            active_deadline_seconds=(
                override.max_step_duration_in_seconds
                if self.max_step_duration_in_seconds != override.max_step_duration_in_seconds
                else None
            ),
            container=container,
            retry_strategy=override.retry_strategy if self.retry_strategy != override.retry_strategy else None,
            tolerations=override.tolerations if self.tolerations != override.tolerations else None,
            node_selector=override.node_selector if self.node_selector != override.node_selector else None,
        )

        # inputs are the "iterate_as" value map variables in the same order as they are observed
        # We need to expose the map variables in the command of the container
        if inputs:
            if not container_template.inputs:
                container_template.inputs = []
            container_template.inputs.extend(inputs)

        # The map step fan out would create an output that we should propagate via Argo
        if outputs:
            if not container_template.outputs:
                container_template.outputs = []
            container_template.outputs.extend(outputs)

        return container_template

    def _create_fan_out_template(self, composite_node, list_of_iter_values: Optional[List] = None):
        clean_name = self.get_clean_name(composite_node)
        inputs = []
        # If we are fanning out already map state, we need to send the map variable inside
        # The container template also should be accepting an input parameter
        map_variable = None
        if list_of_iter_values:
            map_variable = self.compose_map_variable(list_of_iter_values=list_of_iter_values)

            for val in list_of_iter_values:
                inputs.append(Parameter(name=val))

        command = utils.get_fan_command(
            mode="out",
            node=composite_node,
            run_id=self._run_id_placeholder,
            map_variable=map_variable,
        )

        outputs = []
        # If the node is a map node, we have to set the output parameters
        # Output is always the step's internal name + iterate-on
        if composite_node.node_type == "map":
            output_parameter = OutputParameter(name="iterate-on")
            outputs.append(output_parameter)

        # Create the node now
        step_config = {"command": command, "type": "task", "next": "dummy"}
        node = create_node(name=f"{clean_name}-fan-out", step_config=step_config)

        container_template = self.create_container_template(
            working_on=node,
            command=command,
            outputs=outputs,
            inputs=inputs,
            overwrite_name=f"{clean_name}-fan-out",
        )

        self._container_templates.append(container_template)
        return DagTaskTemplate(name=f"{clean_name}-fan-out", template=f"{clean_name}-fan-out")

    def _create_fan_in_template(self, composite_node, list_of_iter_values: Optional[List] = None):
        clean_name = self.get_clean_name(composite_node)
        inputs = []
        # If we are fanning in already map state, we need to send the map variable inside
        # The container template also should be accepting an input parameter
        map_variable = None
        if list_of_iter_values:
            map_variable = self.compose_map_variable(list_of_iter_values=list_of_iter_values)

            for val in list_of_iter_values:
                inputs.append(Parameter(name=val))

        command = utils.get_fan_command(
            mode="in",
            node=composite_node,
            run_id=self._run_id_placeholder,
            map_variable=map_variable,
        )

        step_config = {"command": command, "type": "task", "next": "dummy"}
        node = create_node(name=f"{clean_name}-fan-in", step_config=step_config)
        container_template = self.create_container_template(
            working_on=node,
            command=command,
            inputs=inputs,
            overwrite_name=f"{clean_name}-fan-in",
        )
        self._container_templates.append(container_template)
        clean_name = self.get_clean_name(composite_node)
        return DagTaskTemplate(name=f"{clean_name}-fan-in", template=f"{clean_name}-fan-in")

    def _gather_task_templates_of_dag(
        self, dag: Graph, dag_name="magnus-dag", list_of_iter_values: Optional[List] = None
    ):
        current_node = dag.start_at
        previous_node = None
        previous_node_template_name = None

        templates: Dict[str, DagTaskTemplate] = {}

        if not list_of_iter_values:
            list_of_iter_values = []

        while True:
            working_on = dag.get_node_by_name(current_node)
            if previous_node == current_node:
                raise Exception("Potentially running in a infinite loop")

            render_obj = get_renderer(working_on)(executor=self, node=working_on)
            render_obj.render(list_of_iter_values=list_of_iter_values.copy())

            clean_name = self.get_clean_name(working_on)

            # If a task template for clean name exists, retrieve it (could have been created by on_failure)
            template = templates.get(clean_name, DagTaskTemplate(name=clean_name, template=clean_name))

            # Link the current node to previous node, if the previous node was successful.
            if previous_node:
                template.depends.append(f"{previous_node_template_name}.Succeeded")

            templates[clean_name] = template

            # On failure nodes
            if working_on.node_type not in ["success", "fail"] and working_on._get_on_failure_node():
                failure_node = dag.get_node_by_name(working_on._get_on_failure_node())

                failure_template_name = self.get_clean_name(failure_node)
                # If a task template for clean name exists, retrieve it
                failure_template = templates.get(
                    failure_template_name,
                    DagTaskTemplate(name=failure_template_name, template=failure_template_name),
                )
                failure_template.depends.append(f"{clean_name}.Failed")

                templates[failure_template_name] = failure_template

            # If we are in a map node, we need to add the values as arguments
            template = templates[clean_name]
            if list_of_iter_values:
                if not template.arguments:
                    template.arguments = []
                for value in list_of_iter_values:
                    template.arguments.append(Argument(name=value, value="{{inputs.parameters." + value + "}}"))

            # Move ahead to the next node
            previous_node = current_node
            previous_node_template_name = self.get_clean_name(working_on)

            if working_on.node_type in ["success", "fail"]:
                break

            current_node = working_on._get_next_node()

        # Add the iteration values as input to dag template
        dag_template = DagTemplate(tasks=list(templates.values()), name=dag_name)
        if list_of_iter_values:
            if not dag_template.inputs:
                dag_template.inputs = []
            dag_template.inputs.extend([Parameter(name=val) for val in list_of_iter_values])

        # Add the dag template to the list of templates
        self._dag_templates.append(dag_template)

    def _get_template_defaults(self) -> TemplateDefaults:
        user_provided_config = self.model_dump(by_alias=False)

        return TemplateDefaults(**user_provided_config)

    def execute_graph(self, dag: Graph, map_variable: Optional[dict] = None, **kwargs):
        # TODO: Add metadata
        arguments = []
        # Expose "simple" parameters as workflow arguments for dynamic behavior
        if self.expose_parameters_as_inputs:
            for key, value in self._get_parameters().items():
                if isinstance(value, dict) or isinstance(value, list):
                    continue
                env_var = EnvVar(name=key, value=value)
                arguments.append(env_var)

        run_id_var = EnvVar(name="run_id", value="{{workflow.uid}}")
        arguments.append(run_id_var)

        # # TODO: Experimental feature

        # original_run_id_var = EnvVar(name="original_run_id")
        # arguments.append(original_run_id_var)

        for volume in self.spec.volumes:
            self._container_volumes.append(ContainerVolume(name=volume.name, mount_path=volume.mount_path))

        # Container specifications are globally collected and added at the end.
        # Dag specifications are added as part of the dag traversal.
        templates: List[Union[DagTemplate, ContainerTemplate]] = []
        self._gather_task_templates_of_dag(dag=dag, list_of_iter_values=[])
        templates.extend(self._dag_templates)
        templates.extend(self._container_templates)

        spec = self.spec
        spec.templates = templates
        spec.arguments = arguments
        workflow = Workflow(metadata=self.metadata, spec=spec)

        yaml = YAML()
        with open(self.output_file, "w") as f:
            yaml.indent(mapping=2, sequence=4, offset=2)

            yaml.dump(workflow.model_dump(by_alias=True, exclude_none=True), f)

    def execute_job(self, node: BaseNode):
        """
        Use K8's job instead
        """
        raise NotImplementedError("Use K8's job instead")

    def send_return_code(self, stage="traversal"):
        """
        Convenience function used by pipeline to send return code to the caller of the cli

        Raises:
            Exception: If the pipeline execution failed
        """
        if stage != "traversal":  # traversal does no actual execution, so return code is pointless
            run_id = self._context.run_id

            run_log = self._context.run_log_store.get_run_log_by_id(run_id=run_id, full=False)
            if run_log.status == defaults.FAIL:
                raise exceptions.ExecutionFailedError(run_id)


class FileSystemRunLogStore(BaseIntegration):
    """
    Only local execution mode is possible for Buffered Run Log store
    """

    executor_type = "argo"
    service_type = "run_log_store"  # One of secret, catalog, datastore
    service_provider = "file-system"  # The actual implementation of the service

    def validate(self, **kwargs):
        msg = (
            "Argo cannot run work with file-system run log store. "
            "Unless you have made a mechanism to use volume mounts."
            "Using this run log store if the pipeline has concurrent tasks might lead to unexpected results"
        )
        logger.warning(msg)


class ChunkedFileSystemRunLogStore(BaseIntegration):
    """
    Only local execution mode is possible for Buffered Run Log store
    """

    executor_type = "argo"
    service_type = "run_log_store"  # One of secret, catalog, datastore
    service_provider = "chunked-fs"  # The actual implementation of the service

    def validate(self, **kwargs):
        msg = (
            "Argo cannot run work with chunked file-system run log store. "
            "Unless you have made a mechanism to use volume mounts"
        )
        logger.warning(msg)


class FileSystemCatalog(BaseIntegration):
    """
    Only local execution mode is possible for Buffered Run Log store
    """

    executor_type = "argo"
    service_type = "catalog"  # One of secret, catalog, datastore
    service_provider = "file-system"  # The actual implementation of the service

    def validate(self, **kwargs):
        msg = (
            "Argo cannot run work with file-system run log store. Unless you have made a mechanism to use volume mounts"
        )
        logger.warning(msg)



import datetime
import json
import logging
from pathlib import Path
from string import Template
from typing import Any, Dict, List, Optional, Union, cast

from magnus import defaults, utils
from magnus.extensions.run_log_store.generic_chunked import ChunkedRunLogStore

logger = logging.getLogger(defaults.LOGGER_NAME)


class DBRunLogStore(ChunkedRunLogStore):
    """
    File system run log store but chunks the run log into thread safe chunks.
    This enables executions to be parallel.
    """

    service_name: str = "chunked-fs"
    connection_string: str
    db_name: str

    _DB_LOG: Any = None
    _engine: Any = None
    _session: Any = None
    _connection_string: str = ""
    _base: Any = None

    def model_post_init(self, _: Any) -> None:
        run_context = self._context

        secrets = cast(Dict[str, str], run_context.secrets_handler.get())
        connection_string = Template(self.connection_string).safe_substitute(**secrets)

        try:
            import sqlalchemy
            from sqlalchemy import Column, DateTime, Integer, Sequence, Text
            from sqlalchemy.orm import declarative_base, sessionmaker

            Base = declarative_base()

            class DBLog(Base):
                """
                Base table for storing run logs in database.

                In this model, we fragment the run log into logical units that are concurrent safe.
                """

                __tablename__ = self.db_name
                pk = Column(Integer, Sequence("id_seq"), primary_key=True)
                run_id = Column(Text, index=True)
                attribute_key = Column(Text)  # run_log, step_internal_name, parameter_key etc
                attribute_type = Column(Text)  # RunLog, Step, Branch, Parameter
                attribute_value = Column(Text)  # The JSON string
                created_at = Column(DateTime, default=datetime.datetime.utcnow)

            self._engine = sqlalchemy.create_engine(connection_string, pool_pre_ping=True)
            self._session = sessionmaker(bind=self._engine)
            self._DB_LOG = DBLog
            self._connection_string = connection_string
            self._base = Base

        except ImportError as _e:
            logger.exception("Unable to import SQLalchemy, is it installed?")
            msg = "SQLAlchemy is required for this extension. Please install it"
            raise Exception(msg) from _e

    def create_tables(self):
        import sqlalchemy

        engine = sqlalchemy.create_engine(self._connection_string)
        self._base.metadata.create_all(engine)

    def get_matches(self, run_id: str, name: str, multiple_allowed: bool = False) -> Optional[Union[List[Path], Path]]:
        """
        Get contents of files matching the pattern name*

        Args:
            run_id (str): The run id
            name (str): The suffix of the file name to check in the run log store.
        """
        log_folder = self.log_folder_with_run_id(run_id=run_id)

        sub_name = Template(name).safe_substitute({"creation_time": ""})

        matches = list(log_folder.glob(f"{sub_name}*"))
        if matches:
            if not multiple_allowed:
                if len(matches) > 1:
                    msg = f"Multiple matches found for {name} while multiple is not allowed"
                    raise Exception(msg)
                return matches[0]
            return matches

        return None

    def log_folder_with_run_id(self, run_id: str) -> Path:
        """
        Utility function to get the log folder for a run id.

        Args:
            run_id (str): The run id

        Returns:
            Path: The path to the log folder with the run id
        """
        return Path(self.log_folder) / run_id

    def safe_suffix_json(self, name: Union[Path, str]) -> str:
        """
        Safely attach a suffix to a json file.

        Args:
            name (Path): The name of the file with or without suffix of json

        Returns:
            str : The name of the file with .json
        """
        if str(name).endswith("json"):
            return str(name)

        return str(name) + ".json"

    def _store(self, run_id: str, contents: dict, name: Union[Path, str], insert=False):
        """
        Store the contents against the name in the folder.

        Args:
            run_id (str): The run id
            contents (dict): The dict to store
            name (str): The name to store as
        """
        if insert:
            name = self.log_folder_with_run_id(run_id=run_id) / name

        utils.safe_make_dir(self.log_folder_with_run_id(run_id=run_id))

        with open(self.safe_suffix_json(name), "w") as fw:
            json.dump(contents, fw, ensure_ascii=True, indent=4)

    def _retrieve(self, name: Path) -> dict:
        """
        Does the job of retrieving from the folder.

        Args:
            name (str): the name of the file to retrieve

        Returns:
            dict: The contents
        """
        contents: dict = {}

        with open(self.safe_suffix_json(name), "r") as fr:
            contents = json.load(fr)

        return contents



import ast
import contextlib
import importlib
import io
import json
import logging
import os
import subprocess
import sys
from typing import Any, Dict, Tuple

from pydantic import BaseModel, ConfigDict, Field, ValidationInfo, field_validator
from pydantic._internal._model_construction import ModelMetaclass
from stevedore import driver

import magnus.context as context
from magnus import defaults, parameters, utils
from magnus.defaults import TypeMapVariable

logger = logging.getLogger(defaults.LOGGER_NAME)
logging.getLogger("stevedore").setLevel(logging.CRITICAL)


# TODO: Can we add memory peak, cpu usage, etc. to the metrics?


class BaseTaskType(BaseModel):
    """A base task class which does the execution of command defined by the user."""

    task_type: str = Field(serialization_alias="command_type")
    node_name: str = Field(exclude=True)

    model_config = ConfigDict(extra="forbid")

    @property
    def _context(self):
        return context.run_context

    def get_cli_options(self) -> Tuple[str, dict]:
        """
        Key is the name of the cli option and value is the value of the cli option.
        This should always be in sync with the cli options defined in execute_*.

        Returns:
            str: The name of the cli option.
            dict: The dict of cli options for the task.

        Raises:
            NotImplementedError: Base class, not implemented
        """
        raise NotImplementedError()

    def _get_parameters(self, map_variable: TypeMapVariable = None, **kwargs) -> Dict[str, Any]:
        """
        By this step, all the parameters are present as environment variables as json strings.
        Return the parameters in scope for the execution.

        Args:
            map_variable (dict, optional): If the command is part of map node, the value of map. Defaults to None.

        Returns:
            dict: The parameters dictionary in-scope for the task execution
        """
        return parameters.get_user_set_parameters(remove=False)

    def execute_command(self, map_variable: TypeMapVariable = None, **kwargs):
        """The function to execute the command.

        And map_variable is sent in as an argument into the function.

        Args:
            map_variable (dict, optional): If the command is part of map node, the value of map. Defaults to None.

        Raises:
            NotImplementedError: Base class, not implemented
        """
        raise NotImplementedError()

    def _set_parameters(self, params: BaseModel, **kwargs):
        """Set the parameters back to the environment variables.

        Args:
            parameters (dict, optional): The parameters to set back as env variables. Defaults to None.
        """
        # Nothing to do
        if not params:
            return

        if not isinstance(params, BaseModel) or isinstance(params, ModelMetaclass):
            raise ValueError("Output variable of a function can only be a pydantic model or dynamic model.")

        parameters.set_user_defined_params_as_environment_variables(params.model_dump(by_alias=True))

    @contextlib.contextmanager
    def output_to_file(self, map_variable: TypeMapVariable = None):
        """Context manager to put the output of a function execution to catalog.

        Args:
            map_variable (dict, optional): If the command is part of map node, the value of map. Defaults to None.

        """
        from magnus import put_in_catalog  # Causing cyclic imports

        log_file_name = self.node_name.replace(" ", "_") + ".execution.log"
        if map_variable:
            for _, value in map_variable.items():
                log_file_name += "_" + str(value)

        log_file = open(log_file_name, "w")

        f = io.StringIO()
        try:
            with contextlib.redirect_stdout(f):
                yield
        finally:
            print(f.getvalue())  # print to console
            log_file.write(f.getvalue())  # Print to file

            f.close()
            log_file.close()
            put_in_catalog(log_file.name)
            os.remove(log_file.name)


class EasyModel(BaseModel):
    model_config = ConfigDict(extra="allow")


def make_pydantic_model(
    variables: Dict[str, Any],
    prefix: str = "",
) -> BaseModel:
    prefix_removed = {utils.remove_prefix(k, prefix): v for k, v in variables.items()}
    return EasyModel(**prefix_removed)


class PythonTaskType(BaseTaskType):  # pylint: disable=too-few-public-methods
    """The task class for python command."""

    task_type: str = Field(default="python", serialization_alias="command_type")
    command: str

    @field_validator("command")
    @classmethod
    def validate_command(cls, command: str):
        if not command:
            raise Exception("Command cannot be empty for shell task")

        return command

    def get_cli_options(self) -> Tuple[str, dict]:
        """Return the cli options for the task.

        Returns:
            dict: The cli options for the task
        """
        return "function", {"command": self.command}

    def execute_command(self, map_variable: TypeMapVariable = None, **kwargs):
        """Execute the notebook as defined by the command."""
        module, func = utils.get_module_and_attr_names(self.command)
        sys.path.insert(0, os.getcwd())  # Need to add the current directory to path
        imported_module = importlib.import_module(module)
        f = getattr(imported_module, func)

        params = self._get_parameters()
        filtered_parameters = parameters.filter_arguments_for_func(f, params, map_variable)

        if map_variable:
            os.environ[defaults.MAP_VARIABLE] = json.dumps(map_variable)

        logger.info(f"Calling {func} from {module} with {filtered_parameters}")

        with self.output_to_file(map_variable=map_variable) as _:
            try:
                user_set_parameters = f(**filtered_parameters)
            except Exception as _e:
                msg = f"Call to the function {self.command} with {filtered_parameters} did not succeed.\n"
                logger.exception(msg)
                logger.exception(_e)
                raise

            if map_variable:
                del os.environ[defaults.MAP_VARIABLE]

            self._set_parameters(user_set_parameters)


class NotebookTaskType(BaseTaskType):
    """The task class for Notebook based execution."""

    task_type: str = Field(default="notebook", serialization_alias="command_type")
    command: str
    notebook_output_path: str = Field(default="", validate_default=True)
    output_cell_tag: str = Field(default="magnus_output", validate_default=True)
    optional_ploomber_args: dict = {}

    _output_tag: str = "magnus_output"

    @field_validator("command")
    @classmethod
    def notebook_should_end_with_ipynb(cls, command: str):
        if not command:
            raise Exception("Command should point to the ipynb file")

        if not command.endswith(".ipynb"):
            raise Exception("Notebook task should point to a ipynb file")

        return command

    @field_validator("notebook_output_path")
    @classmethod
    def correct_notebook_output_path(cls, notebook_output_path: str, info: ValidationInfo):
        if notebook_output_path:
            return notebook_output_path

        command = info.data["command"]
        return "".join(command.split(".")[:-1]) + "_out.ipynb"

    def get_cli_options(self) -> Tuple[str, dict]:
        return "notebook", {"command": self.command, "notebook-output-path": self.notebook_output_path}

    def _parse_notebook_for_output(self, notebook: Any):
        collected_params = {}

        for cell in notebook.cells:
            d = cell.dict()
            # identify the tags attached to the cell.
            tags = d.get("metadata", {}).get("tags", {})
            if self.output_cell_tag in tags:
                # There is a tag that has output
                outputs = d["outputs"]

                for out in outputs:
                    params = out.get("text", "{}")
                    collected_params.update(ast.literal_eval(params))

        return collected_params

    def execute_command(self, map_variable: TypeMapVariable = None, **kwargs):
        """Execute the python notebook as defined by the command.

        Args:
            map_variable (dict, optional): If the node is part of internal branch. Defaults to None.

        Raises:
            ImportError: If necessary dependencies are not installed
            Exception: If anything else fails
        """
        try:
            import ploomber_engine as pm

            from magnus import put_in_catalog  # Causes issues with cyclic import

            parameters = self._get_parameters()
            filtered_parameters = parameters

            notebook_output_path = self.notebook_output_path

            if map_variable:
                os.environ[defaults.MAP_VARIABLE] = json.dumps(map_variable)

                for _, value in map_variable.items():
                    notebook_output_path += "_" + str(value)

            ploomber_optional_args = self.optional_ploomber_args

            kwds = {
                "input_path": self.command,
                "output_path": notebook_output_path,
                "parameters": filtered_parameters,
                "log_output": True,
                "progress_bar": False,
            }
            kwds.update(ploomber_optional_args)

            collected_params: Dict[str, Any] = {}
            with self.output_to_file(map_variable=map_variable) as _:
                out = pm.execute_notebook(**kwds)
                collected_params = self._parse_notebook_for_output(out)

            collected_params_model = make_pydantic_model(collected_params)
            self._set_parameters(collected_params_model)

            put_in_catalog(notebook_output_path)
            if map_variable:
                del os.environ[defaults.MAP_VARIABLE]

        except ImportError as e:
            msg = (
                "Task type of notebook requires ploomber engine to be installed. Please install via optional: notebook"
            )
            raise Exception(msg) from e


class ShellTaskType(BaseTaskType):
    """
    The task class for shell based commands.
    """

    task_type: str = Field(default="shell", serialization_alias="command_type")
    command: str

    @field_validator("command")
    @classmethod
    def validate_command(cls, command: str):
        if not command:
            raise Exception("Command cannot be empty for shell task")

        return command

    def execute_command(self, map_variable: TypeMapVariable = None, **kwargs):
        # Using shell=True as we want to have chained commands to be executed in the same shell.
        """Execute the shell command as defined by the command.

        Args:
            map_variable (dict, optional): If the node is part of an internal branch. Defaults to None.
        """
        subprocess_env = os.environ.copy()

        if map_variable:
            subprocess_env[defaults.MAP_VARIABLE] = json.dumps(map_variable)

        command = self.command.strip() + " && env | grep MAGNUS"
        logger.info(f"Executing shell command: {command}")

        output_parameters = {}

        with subprocess.Popen(
            command,
            shell=True,
            env=subprocess_env,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        ) as proc, self.output_to_file(map_variable=map_variable) as _:
            for line in proc.stdout:  # type: ignore
                logger.info(line)
                print(line)

                if line.startswith(defaults.PARAMETER_PREFIX):
                    key, value = line.strip().split("=", 1)
                    try:
                        output_parameters[key] = json.loads(value)
                    except json.JSONDecodeError:
                        output_parameters[key] = value  # simple data types

                if line.startswith(defaults.TRACK_PREFIX):
                    key, value = line.split("=", 1)
                    os.environ[key] = value.strip()

            proc.wait()
            if proc.returncode != 0:
                raise Exception("Command failed")

        self._set_parameters(
            params=make_pydantic_model(
                output_parameters,
                defaults.PARAMETER_PREFIX,
            )
        )


def create_task(kwargs_for_init) -> BaseTaskType:
    """
    Creates a task object from the command configuration.

    Args:
        A dictionary of keyword arguments that are sent by the user to the task.
        Check against the model class for the validity of it.

    Returns:
        tasks.BaseTaskType: The command object
    """
    # The dictionary cannot be modified
    kwargs = kwargs_for_init.copy()
    command_type = kwargs.pop("command_type", defaults.COMMAND_TYPE)

    try:
        task_mgr = driver.DriverManager(
            namespace="tasks",
            name=command_type,
            invoke_on_load=True,
            invoke_kwds=kwargs,
        )
        return task_mgr.driver
    except Exception as _e:
        msg = (
            f"Could not find the task type {command_type}. Please ensure you have installed "
            "the extension that provides the node type."
        )
        raise Exception(msg) from _e



import json
import logging
from pathlib import Path
from string import Template
from typing import Optional, Sequence, Union

from magnus import defaults, utils
from magnus.extensions.run_log_store.generic_chunked import ChunkedRunLogStore

logger = logging.getLogger(defaults.LOGGER_NAME)

T = Union[str, Path]


class ChunkedFileSystemRunLogStore(ChunkedRunLogStore):
    """
    File system run log store but chunks the run log into thread safe chunks.
    This enables executions to be parallel.
    """

    service_name: str = "chunked-fs"
    log_folder: str = defaults.LOG_LOCATION_FOLDER

    def get_matches(self, run_id: str, name: str, multiple_allowed: bool = False) -> Optional[Union[Sequence[T], T]]:
        """
        Get contents of files matching the pattern name*

        Args:
            run_id (str): The run id
            name (str): The suffix of the file name to check in the run log store.
        """
        log_folder = self.log_folder_with_run_id(run_id=run_id)

        sub_name = Template(name).safe_substitute({"creation_time": ""})

        matches = list(log_folder.glob(f"{sub_name}*"))
        if matches:
            if not multiple_allowed:
                if len(matches) > 1:
                    msg = f"Multiple matches found for {name} while multiple is not allowed"
                    raise Exception(msg)
                return matches[0]
            return matches

        return None

    def log_folder_with_run_id(self, run_id: str) -> Path:
        """
        Utility function to get the log folder for a run id.

        Args:
            run_id (str): The run id

        Returns:
            Path: The path to the log folder with the run id
        """
        return Path(self.log_folder) / run_id

    def safe_suffix_json(self, name: Union[Path, str]) -> str:
        """
        Safely attach a suffix to a json file.

        Args:
            name (Path): The name of the file with or without suffix of json

        Returns:
            str : The name of the file with .json
        """
        if str(name).endswith("json"):
            return str(name)

        return str(name) + ".json"

    def _store(self, run_id: str, contents: dict, name: Union[Path, str], insert=False):
        """
        Store the contents against the name in the folder.

        Args:
            run_id (str): The run id
            contents (dict): The dict to store
            name (str): The name to store as
        """
        if insert:
            name = self.log_folder_with_run_id(run_id=run_id) / name

        utils.safe_make_dir(self.log_folder_with_run_id(run_id=run_id))

        with open(self.safe_suffix_json(name), "w") as fw:
            json.dump(contents, fw, ensure_ascii=True, indent=4)

    def _retrieve(self, name: Union[str, Path]) -> dict:
        """
        Does the job of retrieving from the folder.

        Args:
            name (str): the name of the file to retrieve

        Returns:
            dict: The contents
        """
        contents: dict = {}

        with open(self.safe_suffix_json(name), "r") as fr:
            contents = json.load(fr)

        return contents



import logging
import time
from abc import abstractmethod
from enum import Enum
from pathlib import Path
from string import Template
from typing import Any, Dict, Optional, Sequence, Union

from magnus import defaults, exceptions
from magnus.datastore import BaseRunLogStore, BranchLog, RunLog, StepLog

logger = logging.getLogger(defaults.LOGGER_NAME)


T = Union[str, Path]  # Holds str, path


class EntityNotFoundError(Exception):
    pass


class ChunkedRunLogStore(BaseRunLogStore):
    """
    A generic implementation of a RunLogStore that stores RunLogs in chunks.
    """

    service_name: str = ""

    class LogTypes(Enum):
        RUN_LOG = "RunLog"
        PARAMETER = "Parameter"
        STEP_LOG = "StepLog"
        BRANCH_LOG = "BranchLog"

    class ModelTypes(Enum):
        RUN_LOG = RunLog
        PARAMETER = dict
        STEP_LOG = StepLog
        BRANCH_LOG = BranchLog

    def naming_pattern(self, log_type: LogTypes, name: str = "") -> str:
        """
        Naming pattern to store RunLog, Parameter, StepLog or BranchLog.

        The reasoning for name to be defaulted to empty string:
            Its actually conditionally empty. For RunLog and Parameter it is empty.
            For StepLog and BranchLog it should be provided.

        Args:
            log_type (LogTypes): One of RunLog, Parameter, StepLog or BranchLog
            name (str, optional): The name to be included or left. Defaults to ''.

        Raises:
            Exception: If log_type is not recognized

        Returns:
            str: The naming pattern
        """
        if log_type == self.LogTypes.RUN_LOG:
            return f"{self.LogTypes.RUN_LOG.value}"

        if log_type == self.LogTypes.PARAMETER:
            return "-".join([self.LogTypes.PARAMETER.value, name])

        if not name:
            raise Exception(f"Name should be provided for naming pattern for {log_type}")

        if log_type == self.LogTypes.STEP_LOG:
            return "-".join([self.LogTypes.STEP_LOG.value, name, "${creation_time}"])

        if log_type == self.LogTypes.BRANCH_LOG:
            return "-".join([self.LogTypes.BRANCH_LOG.value, name, "${creation_time}"])

        raise Exception("Unexpected log type")

    @abstractmethod
    def get_matches(self, run_id: str, name: str, multiple_allowed: bool = False) -> Optional[Union[Sequence[T], T]]:
        """
        Get contents of persistence layer matching the pattern name*

        Args:
            run_id (str): The run id
            name (str): The suffix of the entity name to check in the run log store.
        """
        ...

    @abstractmethod
    def _store(self, run_id: str, contents: dict, name: T, insert: bool = False):
        """
        Store the contents against the name in the persistence layer.

        Args:
            run_id (str): The run id
            contents (dict): The dict to store
            name (str): The name to store as
        """
        ...

    @abstractmethod
    def _retrieve(self, name: T) -> dict:
        """
        Does the job of retrieving from the persistent layer.

        Args:
            name (str): the name of the file to retrieve

        Returns:
            dict: The contents
        """
        ...

    def store(self, run_id: str, log_type: LogTypes, contents: dict, name: str = ""):
        """Store a SINGLE log type in the file system

        Args:
            run_id (str): The run id to store against
            log_type (LogTypes): The type of log to store
            contents (dict): The dict of contents to store
            name (str, optional): The name against the contents have to be stored. Defaults to ''.
        """
        naming_pattern = self.naming_pattern(log_type=log_type, name=name)
        match = self.get_matches(run_id=run_id, name=naming_pattern, multiple_allowed=False)
        # The boolean multiple allowed confuses mypy a lot!
        name_to_give: str = ""
        insert = False

        if match:
            existing_contents = self._retrieve(name=match)  # type: ignore
            contents = dict(existing_contents, **contents)
            name_to_give = match  # type: ignore
        else:
            name_to_give = Template(naming_pattern).safe_substitute({"creation_time": str(int(time.time_ns()))})
            insert = True

        self._store(run_id=run_id, contents=contents, name=name_to_give, insert=insert)

    def retrieve(self, run_id: str, log_type: LogTypes, name: str = "", multiple_allowed=False) -> Any:
        """
        Retrieve the model given a log_type and a name.
        Use multiple_allowed to control if you are expecting multiple of them.
        eg: There could be multiple of Parameters- but only one of StepLog-stepname

        The reasoning for name to be defaulted to empty string:
            Its actually conditionally empty. For RunLog and Parameter it is empty.
            For StepLog and BranchLog it should be provided.

        Args:
            run_id (str): The run id
            log_type (LogTypes): One of RunLog, Parameter, StepLog, BranchLog
            name (str, optional): The name to match. Defaults to ''.
            multiple_allowed (bool, optional): Are multiple allowed. Defaults to False.

        Raises:
            FileNotFoundError: If there is no match found

        Returns:
            Any: One of StepLog, BranchLog, Parameter or RunLog
        """
        # The reason of any is it could be one of Logs or dict or list of the
        if not name and log_type not in [
            self.LogTypes.RUN_LOG,
            self.LogTypes.PARAMETER,
        ]:
            raise Exception(f"Name is required during retrieval for {log_type}")

        naming_pattern = self.naming_pattern(log_type=log_type, name=name)
        matches = self.get_matches(run_id=run_id, name=naming_pattern, multiple_allowed=multiple_allowed)
        if matches:
            if not multiple_allowed:
                contents = self._retrieve(name=matches)  # type: ignore
                model = self.ModelTypes[log_type.name].value
                return model(**contents)

            models = []
            for match in matches:  # type: ignore
                contents = self._retrieve(name=match)
                model = self.ModelTypes[log_type.name].value
                models.append(model(**contents))
            return models

        raise EntityNotFoundError()

    def orderly_retrieve(self, run_id: str, log_type: LogTypes) -> Dict[str, Union[StepLog, BranchLog]]:
        """Should only be used by prepare full run log.

        Retrieves the StepLog or BranchLog sorted according to creation time.

        Args:
            run_id (str): _description_
            log_type (LogTypes): _description_
        """
        prefix: str = self.LogTypes.STEP_LOG.value

        if log_type == self.LogTypes.BRANCH_LOG:
            prefix = self.LogTypes.BRANCH_LOG.value

        matches = self.get_matches(run_id=run_id, name=prefix, multiple_allowed=True)

        if log_type == self.LogTypes.BRANCH_LOG and not matches:
            # No branch logs are found
            return {}
        # Forcing get_matches to always return a list is a better design
        epoch_created = [str(match).split("-")[-1] for match in matches]  # type: ignore

        # sort matches by epoch created
        epoch_created, matches = zip(*sorted(zip(epoch_created, matches)))  # type: ignore

        logs: Dict[str, Union[StepLog, BranchLog]] = {}

        for match in matches:
            model = self.ModelTypes[log_type.name].value
            log_model = model(**self._retrieve(match))
            logs[log_model.internal_name] = log_model  # type: ignore

        return logs

    def _get_parent_branch(self, name: str) -> Union[str, None]:
        """
        Returns the name of the parent branch.
        If the step is part of main dag, return None.

        Args:
            name (str): The name of the step.

        Returns:
            str: The name of the branch containing the step.
        """
        dot_path = name.split(".")

        if len(dot_path) == 1:
            return None
        # Ignore the step name
        return ".".join(dot_path[:-1])

    def _get_parent_step(self, name: str) -> Union[str, None]:
        """
        Returns the step containing the step, useful when we have steps within a branch.
        Returns None, if the step belongs to parent dag.

        Args:
            name (str): The name of the step to find the parent step it belongs to.

        Returns:
            str: The parent step the step belongs to, None if the step belongs to parent dag.
        """
        dot_path = name.split(".")

        if len(dot_path) == 1:
            return None
        # Ignore the branch.step_name
        return ".".join(dot_path[:-2])

    def _prepare_full_run_log(self, run_log: RunLog):
        """
        Populates the run log with the branches and steps.

        Args:
            run_log (RunLog): The partial run log containing empty step logs
        """
        run_id = run_log.run_id
        run_log.parameters = self.get_parameters(run_id=run_id)

        ordered_steps = self.orderly_retrieve(run_id=run_id, log_type=self.LogTypes.STEP_LOG)
        ordered_branches = self.orderly_retrieve(run_id=run_id, log_type=self.LogTypes.BRANCH_LOG)

        current_branch: Any = None  # It could be str, None, RunLog
        for step_internal_name in ordered_steps:
            current_branch = self._get_parent_branch(step_internal_name)
            step_to_add_branch = self._get_parent_step(step_internal_name)

            if not current_branch:
                current_branch = run_log
            else:
                current_branch = ordered_branches[current_branch]
                step_to_add_branch = ordered_steps[step_to_add_branch]  # type: ignore
                step_to_add_branch.branches[current_branch.internal_name] = current_branch  # type: ignore

            current_branch.steps[step_internal_name] = ordered_steps[step_internal_name]

    def create_run_log(
        self,
        run_id: str,
        dag_hash: str = "",
        use_cached: bool = False,
        tag: str = "",
        original_run_id: str = "",
        status: str = defaults.CREATED,
        **kwargs,
    ):
        """
        Creates a Run Log object by using the config

        Logically the method should do the following:
            * Creates a Run log
            * Adds it to the db
            * Return the log
        """
        try:
            self.get_run_log_by_id(run_id=run_id, full=False)
            raise exceptions.RunLogExistsError(run_id=run_id)
        except exceptions.RunLogNotFoundError:
            pass

        logger.info(f"{self.service_name} Creating a Run Log for : {run_id}")
        run_log = RunLog(
            run_id=run_id,
            dag_hash=dag_hash,
            use_cached=use_cached,
            tag=tag,
            original_run_id=original_run_id,
            status=status,
        )

        self.store(run_id=run_id, contents=run_log.model_dump(), log_type=self.LogTypes.RUN_LOG)
        return run_log

    def get_run_log_by_id(self, run_id: str, full: bool = False, **kwargs) -> RunLog:
        """
        Retrieves a Run log from the database using the config and the run_id

        Args:
            run_id (str): The run_id of the run
            full (bool): return the full run log store or only the RunLog object

        Returns:
            RunLog: The RunLog object identified by the run_id

        Logically the method should:
            * Returns the run_log defined by id from the data store defined by the config

        """
        try:
            logger.info(f"{self.service_name} Getting a Run Log for : {run_id}")
            run_log = self.retrieve(run_id=run_id, log_type=self.LogTypes.RUN_LOG, multiple_allowed=False)

            if full:
                self._prepare_full_run_log(run_log=run_log)

            return run_log
        except EntityNotFoundError as e:
            raise exceptions.RunLogNotFoundError(run_id) from e

    def put_run_log(self, run_log: RunLog, **kwargs):
        """
        Puts the Run Log in the database as defined by the config

        Args:
            run_log (RunLog): The Run log of the run

        Logically the method should:
            Puts the run_log into the database

        Raises:
            NotImplementedError: This is a base class and therefore has no default implementation
        """
        run_id = run_log.run_id
        self.store(run_id=run_id, contents=run_log.model_dump(), log_type=self.LogTypes.RUN_LOG)

    def get_parameters(self, run_id: str, **kwargs) -> dict:
        """
        Get the parameters from the Run log defined by the run_id

        Args:
            run_id (str): The run_id of the run

        The method should:
            * Call get_run_log_by_id(run_id) to retrieve the run_log
            * Return the parameters as identified in the run_log

        Returns:
            dict: A dictionary of the run_log parameters
        Raises:
            RunLogNotFoundError: If the run log for run_id is not found in the datastore
        """
        parameters = {}
        try:
            parameters_list = self.retrieve(run_id=run_id, log_type=self.LogTypes.PARAMETER, multiple_allowed=True)
            parameters = {key: value for param in parameters_list for key, value in param.items()}
        except EntityNotFoundError:
            # No parameters are set
            pass

        return parameters

    def set_parameters(self, run_id: str, parameters: dict, **kwargs):
        """
        Update the parameters of the Run log with the new parameters

        This method would over-write the parameters, if the parameter exists in the run log already

        The method should:
            * Call get_run_log_by_id(run_id) to retrieve the run_log
            * Update the parameters of the run_log
            * Call put_run_log(run_log) to put the run_log in the datastore

        Args:
            run_id (str): The run_id of the run
            parameters (dict): The parameters to update in the run log
        Raises:
            RunLogNotFoundError: If the run log for run_id is not found in the datastore
        """
        for key, value in parameters.items():
            self.store(
                run_id=run_id,
                log_type=self.LogTypes.PARAMETER,
                contents={key: value},
                name=key,
            )

    def get_run_config(self, run_id: str, **kwargs) -> dict:
        """
        Given a run_id, return the run_config used to perform the run.

        Args:
            run_id (str): The run_id of the run

        Returns:
            dict: The run config used for the run
        """

        run_log = self.get_run_log_by_id(run_id=run_id)
        return run_log.run_config

    def set_run_config(self, run_id: str, run_config: dict, **kwargs):
        """Set the run config used to run the run_id

        Args:
            run_id (str): The run_id of the run
            run_config (dict): The run_config of the run
        """

        run_log = self.get_run_log_by_id(run_id=run_id)
        run_log.run_config.update(run_config)
        self.put_run_log(run_log=run_log)

    def get_step_log(self, internal_name: str, run_id: str, **kwargs) -> StepLog:
        """
        Get a step log from the datastore for run_id and the internal naming of the step log

        The internal naming of the step log is a dot path convention.

        The method should:
            * Call get_run_log_by_id(run_id) to retrieve the run_log
            * Identify the step location by decoding the internal naming
            * Return the step log

        Args:
            internal_name (str): The internal name of the step log
            run_id (str): The run_id of the run

        Returns:
            StepLog: The step log object for the step defined by the internal naming and run_id

        Raises:
            RunLogNotFoundError: If the run log for run_id is not found in the datastore
            StepLogNotFoundError: If the step log for internal_name is not found in the datastore for run_id
        """
        logger.info(f"{self.service_name} Getting the step log: {internal_name} of {run_id}")

        step_log = self.retrieve(
            run_id=run_id,
            log_type=self.LogTypes.STEP_LOG,
            name=internal_name,
            multiple_allowed=False,
        )

        return step_log

    def add_step_log(self, step_log: StepLog, run_id: str, **kwargs):
        """
        Add the step log in the run log as identified by the run_id in the datastore

        The method should:
             * Call get_run_log_by_id(run_id) to retrieve the run_log
             * Identify the branch to add the step by decoding the step_logs internal name
             * Add the step log to the identified branch log
             * Call put_run_log(run_log) to put the run_log in the datastore

        Args:
            step_log (StepLog): The Step log to add to the database
            run_id (str): The run id of the run

        Raises:
            RunLogNotFoundError: If the run log for run_id is not found in the datastore
            BranchLogNotFoundError: If the branch of the step log for internal_name is not found in the datastore
                                    for run_id
        """
        logger.info(f"{self.service_name} Adding the step log to DB: {step_log.internal_name}")

        self.store(
            run_id=run_id,
            log_type=self.LogTypes.STEP_LOG,
            contents=step_log.model_dump(),
            name=step_log.internal_name,
        )

    def get_branch_log(self, internal_branch_name: str, run_id: str, **kwargs) -> Union[BranchLog, RunLog]:
        """
        Returns the branch log by the internal branch name for the run id

        If the internal branch name is none, returns the run log

        Args:
            internal_branch_name (str): The internal branch name to retrieve.
            run_id (str): The run id of interest

        Returns:
            BranchLog: The branch log or the run log as requested.
        """
        if not internal_branch_name:
            return self.get_run_log_by_id(run_id=run_id)
        branch = self.retrieve(run_id=run_id, log_type=self.LogTypes.BRANCH_LOG, name=internal_branch_name)
        return branch

    def add_branch_log(self, branch_log: Union[BranchLog, RunLog], run_id: str, **kwargs):
        """
        The method should:
        # Get the run log
        # Get the branch and step containing the branch
        # Add the branch to the step
        # Write the run_log

        The branch log could some times be a Run log and should be handled appropriately

        Args:
            branch_log (BranchLog): The branch log/run log to add to the database
            run_id (str): The run id to which the branch/run log is added
        """
        if not isinstance(branch_log, BranchLog):
            self.put_run_log(branch_log)
            return

        internal_branch_name = branch_log.internal_name

        logger.info(f"{self.service_name} Adding the branch log to DB: {branch_log.internal_name}")
        self.store(
            run_id=run_id,
            log_type=self.LogTypes.BRANCH_LOG,
            contents=branch_log.model_dump(),
            name=internal_branch_name,
        )



import logging
import os

from magnus import defaults, exceptions, utils
from magnus.secrets import BaseSecrets

logger = logging.getLogger(defaults.LOGGER_NAME)


class DotEnvSecrets(BaseSecrets):
    """
    A secret manager which uses .env files for secrets.

    We recommend this secrets manager only for local development and should not be used for anything close to
    production.
    """

    service_name: str = "dotenv"
    location: str = defaults.DOTENV_FILE_LOCATION
    secrets: dict = {}

    @property
    def secrets_location(self):
        """
        Return the location of the .env file.
        If the user has not over-ridden it, it defaults to .env file in the project root.

        Returns:
            str: The location of the secrets file
        """
        return self.location

    def _load_secrets(self):
        """
        We assume that a dotenv file is of format,
            key=value  -> secrets[key]='value'
            key=value# comment  -> secrets[key1]='value1'
            key=value2 # comment. -> secrets[key2]='value2'

        Any of the above formats with export or set in front of them.

        We strip the secret value of any empty spaces at the start and end.

        Raises:
            Exception: If the file at secrets_location is not found.
            Exception: If the secrets are not formatted correctly.
        """
        # It was loaded in the previous call and need not to be reloaded
        if self.secrets:
            return

        secrets_location = self.secrets_location
        if not utils.does_file_exist(secrets_location):
            raise Exception(f"Did not find the secrets file in {secrets_location}")

        with open(secrets_location, "r") as fr:
            for secret_line in fr:
                # The order of removing fluff around the expression
                # the new line
                # the comment
                # the white space
                # Any export or set in front of the key any spaces after that.

                secret_line = secret_line.strip(os.linesep).split("#")[0].strip()

                if secret_line == "":
                    continue

                secret_line = utils.remove_prefix(secret_line, prefix="export").strip()
                secret_line = utils.remove_prefix(secret_line, prefix="EXPORT").strip()
                secret_line = utils.remove_prefix(secret_line, prefix="set").strip()
                secret_line = utils.remove_prefix(secret_line, prefix="SET").strip()

                data = secret_line.split("=")
                if len(data) != 2:
                    raise Exception("A secret should be of format, secret_name=secret_value[# any comment]")

                key, value = data
                self.secrets[key] = value.strip().strip('"').strip(os.linesep)

    def get(self, name: str = "", **kwargs) -> str:
        """
        Get a secret of name from the secrets file.


        Args:
            name (str): The name of the secret to retrieve

        Raises:
            Exception: If the secret by the name is not found.

        Returns:
            str: The value of the secret
        """
        self._load_secrets()

        if name in self.secrets:
            return self.secrets[name]

        raise exceptions.SecretNotFoundError(secret_name=name, secret_setting=self.secrets_location)



import logging
from pathlib import Path
from typing import Dict, cast

from pydantic import Field
from rich import print

from magnus import defaults, integration, utils
from magnus.datastore import StepLog
from magnus.defaults import TypeMapVariable
from magnus.extensions.executor import GenericExecutor
from magnus.extensions.nodes import TaskNode
from magnus.integration import BaseIntegration
from magnus.nodes import BaseNode

logger = logging.getLogger(defaults.LOGGER_NAME)


class LocalContainerExecutor(GenericExecutor):
    """
    In the mode of local-container, we execute all the commands in a container.

    Ensure that the local compute has enough resources to finish all your jobs.

    The image of the run, could either be provided as default in the configuration of the execution engine
    i.e.:
    execution:
      type: 'local-container'
      config:
        docker_image: the image you want the code to run in.

    or default image could be over-ridden for a single node by providing a docker_image in the step config.
    i.e:
    dag:
      steps:
        step:
          executor_config:
            local-container:
                docker_image: The image that you want that single step to run in.
    This image would only be used for that step only.

    This mode does not build the docker image with the latest code for you, it is still left for the user to build
    and ensure that the docker image provided is the correct one.

    Example config:
    execution:
      type: local-container
      config:
        docker_image: The default docker image to use if the node does not provide one.
    """

    service_name: str = "local-container"
    docker_image: str
    auto_remove_container: bool = True
    run_in_local: bool = False
    environment: Dict[str, str] = Field(default_factory=dict)

    _container_log_location = "/tmp/run_logs/"
    _container_catalog_location = "/tmp/catalog/"
    _container_secrets_location = "/tmp/dotenv"
    _volumes: Dict[str, Dict[str, str]] = {}

    def add_code_identities(self, node: BaseNode, step_log: StepLog, **kwargs):
        """
        Call the Base class to add the git code identity and add docker identity

        Args:
            node (BaseNode): The node we are adding the code identity
            step_log (Object): The step log corresponding to the node
        """

        super().add_code_identities(node, step_log)

        if node.node_type in ["success", "fail"]:
            # Need not add code identities if we are in a success or fail node
            return

        executor_config = self._resolve_executor_config(node)

        docker_image = executor_config.get("docker_image", None)
        if docker_image:
            code_id = self._context.run_log_store.create_code_identity()

            code_id.code_identifier = utils.get_local_docker_image_id(docker_image)
            code_id.code_identifier_type = "docker"
            code_id.code_identifier_dependable = True
            code_id.code_identifier_url = "local docker host"
            step_log.code_identities.append(code_id)

    def execute_node(self, node: BaseNode, map_variable: TypeMapVariable = None, **kwargs):
        """
        We are already in the container, we just execute the node.
        The node is already prepared for execution.
        """
        return self._execute_node(node, map_variable, **kwargs)

    def execute_job(self, node: TaskNode):
        """
        Set up the step log and call the execute node

        Args:
            node (BaseNode): _description_
        """

        step_log = self._context.run_log_store.create_step_log(node.name, node._get_step_log_name(map_variable=None))

        self.add_code_identities(node=node, step_log=step_log)

        step_log.step_type = node.node_type
        step_log.status = defaults.PROCESSING
        self._context.run_log_store.add_step_log(step_log, self._context.run_id)

        command = utils.get_job_execution_command(node)
        self._spin_container(node=node, command=command)

        # Check the step log status and warn if necessary. Docker errors are generally suppressed.
        step_log = self._context.run_log_store.get_step_log(
            node._get_step_log_name(map_variable=None), self._context.run_id
        )
        if step_log.status != defaults.SUCCESS:
            msg = (
                "Node execution inside the container failed. Please check the logs.\n"
                "Note: If you do not see any docker issue from your side and the code works properly on local execution"
                "please raise a bug report."
            )
            logger.warning(msg)

    def trigger_job(self, node: BaseNode, map_variable: TypeMapVariable = None, **kwargs):
        """
        We come into this step via execute from graph, use trigger job to spin up the container.


        If the config has "run_in_local: True", we compute it on local system instead of container.
        In local container execution, we just spin the container to execute magnus execute_single_node.

        Args:
            node (BaseNode): The node we are currently executing
            map_variable (str, optional): If the node is part of the map branch. Defaults to ''.
        """
        executor_config = self._resolve_executor_config(node)
        auto_remove_container = executor_config.get("auto_remove_container", True)

        logger.debug("Here is the resolved executor config")
        logger.debug(executor_config)

        if executor_config.get("run_in_local", False):
            # Do not change config but only validate the configuration.
            # Trigger the job on local system instead of a container
            integration.validate(self, self._context.run_log_store)
            integration.validate(self, self._context.catalog_handler)
            integration.validate(self, self._context.secrets_handler)

            self.execute_node(node=node, map_variable=map_variable, **kwargs)
            return

        command = utils.get_node_execution_command(node, map_variable=map_variable)

        self._spin_container(
            node=node,
            command=command,
            map_variable=map_variable,
            auto_remove_container=auto_remove_container,
            **kwargs,
        )

        step_log = self._context.run_log_store.get_step_log(node._get_step_log_name(map_variable), self._context.run_id)
        if step_log.status != defaults.SUCCESS:
            msg = (
                "Node execution inside the container failed. Please check the logs.\n"
                "Note: If you do not see any docker issue from your side and the code works properly on local execution"
                "please raise a bug report."
            )
            logger.warning(msg)
            step_log.status = defaults.FAIL
            self._context.run_log_store.add_step_log(step_log, self._context.run_id)

    def _spin_container(
        self,
        node: BaseNode,
        command: str,
        map_variable: TypeMapVariable = None,
        auto_remove_container: bool = True,
        **kwargs,
    ):
        """
        During the flow run, we have to spin up a container with the docker image mentioned
        and the right log locations
        """
        # Conditional import
        import docker  # pylint: disable=C0415

        try:
            client = docker.from_env()
            api_client = docker.APIClient()
        except Exception as ex:
            logger.exception("Could not get access to docker")
            raise Exception("Could not get the docker socket file, do you have docker installed?") from ex

        try:
            logger.info(f"Running the command {command}")
            #  Overrides global config with local
            executor_config = self._resolve_executor_config(node)

            docker_image = executor_config.get("docker_image", None)
            environment = executor_config.get("environment", {})
            environment.update(self._context.variables)
            if not docker_image:
                raise Exception(
                    f"Please provide a docker_image using executor_config of the step {node.name} or at global config"
                )

            # TODO: Should consider using getpass.getuser() when running the docker container? Volume permissions
            container = client.containers.create(
                image=docker_image,
                command=command,
                auto_remove=False,
                volumes=self._volumes,
                network_mode="host",
                environment=environment,
            )

            # print(container.__dict__)

            container.start()
            stream = api_client.logs(container=container.id, timestamps=True, stream=True, follow=True)
            while True:
                try:
                    output = next(stream).decode("utf-8")
                    output = output.strip("\r\n")
                    logger.info(output)
                    print(output)
                except StopIteration:
                    logger.info("Docker Run completed")
                    break

            exit_status = api_client.inspect_container(container.id)["State"]["ExitCode"]

            if auto_remove_container:
                container.remove(force=True)

            if exit_status != 0:
                msg = f"Docker command failed with exit code {exit_status}"
                raise Exception(msg)

        except Exception as _e:
            logger.exception("Problems with spinning/running the container")
            raise _e


class LocalContainerComputeFileSystemRunLogstore(BaseIntegration):
    """
    Integration between local container and file system run log store
    """

    executor_type = "local-container"
    service_type = "run_log_store"  # One of secret, catalog, datastore
    service_provider = "file-system"  # The actual implementation of the service

    def validate(self, **kwargs):
        if self.executor._is_parallel_execution():  # pragma: no branch
            msg = (
                "Run log generated by file-system run log store are not thread safe. "
                "Inconsistent results are possible because of race conditions to write to the same file.\n"
                "Consider using partitioned run log store like database for consistent results."
            )
            logger.warning(msg)

    def configure_for_traversal(self, **kwargs):
        from magnus.extensions.run_log_store.file_system.implementation import FileSystemRunLogstore

        self.executor = cast(LocalContainerExecutor, self.executor)
        self.service = cast(FileSystemRunLogstore, self.service)

        write_to = self.service.log_folder_name
        self.executor._volumes[str(Path(write_to).resolve())] = {
            "bind": f"{self.executor._container_log_location}",
            "mode": "rw",
        }

    def configure_for_execution(self, **kwargs):
        from magnus.extensions.run_log_store.file_system.implementation import FileSystemRunLogstore

        self.executor = cast(LocalContainerExecutor, self.executor)
        self.service = cast(FileSystemRunLogstore, self.service)

        self.service.log_folder = self.executor._container_log_location


class LocalContainerComputeFileSystemCatalog(BaseIntegration):
    """
    Integration pattern between Local container and File System catalog
    """

    executor_type = "local-container"
    service_type = "catalog"  # One of secret, catalog, datastore
    service_provider = "file-system"  # The actual implementation of the service

    def configure_for_traversal(self, **kwargs):
        from magnus.extensions.catalog.file_system.implementation import FileSystemCatalog

        self.executor = cast(LocalContainerExecutor, self.executor)
        self.service = cast(FileSystemCatalog, self.service)

        catalog_location = self.service.catalog_location
        self.executor._volumes[str(Path(catalog_location).resolve())] = {
            "bind": f"{self.executor._container_catalog_location}",
            "mode": "rw",
        }

    def configure_for_execution(self, **kwargs):
        from magnus.extensions.catalog.file_system.implementation import FileSystemCatalog

        self.executor = cast(LocalContainerExecutor, self.executor)
        self.service = cast(FileSystemCatalog, self.service)

        self.service.catalog_location = self.executor._container_catalog_location


class LocalContainerComputeDotEnvSecrets(BaseIntegration):
    """
    Integration between local container and dot env secrets
    """

    executor_type = "local-container"
    service_type = "secrets"  # One of secret, catalog, datastore
    service_provider = "dotenv"  # The actual implementation of the service

    def validate(self, **kwargs):
        logger.warning("Using dot env for non local deployments is not ideal, consider options")

    def configure_for_traversal(self, **kwargs):
        from magnus.extensions.secrets.dotenv.implementation import DotEnvSecrets

        self.executor = cast(LocalContainerExecutor, self.executor)
        self.service = cast(DotEnvSecrets, self.service)

        secrets_location = self.service.secrets_location
        self.executor._volumes[str(Path(secrets_location).resolve())] = {
            "bind": f"{self.executor._container_secrets_location}",
            "mode": "ro",
        }

    def configure_for_execution(self, **kwargs):
        from magnus.extensions.secrets.dotenv.implementation import DotEnvSecrets

        self.executor = cast(LocalContainerExecutor, self.executor)
        self.service = cast(DotEnvSecrets, self.service)

        self.service.location = self.executor._container_secrets_location


class LocalContainerComputeEnvSecretsManager(BaseIntegration):
    """
    Integration between local container and env secrets manager
    """

    executor_type = "local-container"
    service_type = "secrets"  # One of secret, catalog, datastore
    service_provider = "env-secrets-manager"  # The actual implementation of the service

    def validate(self, **kwargs):
        msg = (
            "Local container executions cannot be used with environment secrets manager. "
            "Please use a supported secrets manager"
        )
        logger.exception(msg)
        raise Exception(msg)



import logging
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional

from pydantic import BaseModel, ConfigDict, Field, field_validator

import magnus.context as context
from magnus import defaults, exceptions
from magnus.datastore import StepAttempt
from magnus.defaults import TypeMapVariable

logger = logging.getLogger(defaults.LOGGER_NAME)

# --8<-- [start:docs]


class BaseNode(ABC, BaseModel):
    """
    Base class with common functionality provided for a Node of a graph.

    A node of a graph could be a
        * single execution node as task, success, fail.
        * Could be graph in itself as parallel, dag and map.
        * could be a convenience function like as-is.

    The name is relative to the DAG.
    The internal name of the node, is absolute name in dot path convention.
        This has one to one mapping to the name in the run log
    The internal name of a node, should always be odd when split against dot.

    The internal branch name, only applies for branched nodes, is the branch it belongs to.
    The internal branch name should always be even when split against dot.
    """

    node_type: str = Field(serialization_alias="type")
    name: str
    internal_name: str = Field(exclude=True)
    internal_branch_name: str = Field(default="", exclude=True)
    is_composite: bool = Field(default=False, exclude=True)

    @property
    def _context(self):
        return context.run_context

    model_config = ConfigDict(extra="forbid", arbitrary_types_allowed=False)

    @field_validator("name")
    @classmethod
    def validate_name(cls, name: str):
        if "." in name or "%" in name:
            raise ValueError("Node names cannot have . or '%' in them")
        return name

    def _command_friendly_name(self, replace_with=defaults.COMMAND_FRIENDLY_CHARACTER) -> str:
        """
        Replace spaces with special character for spaces.
        Spaces in the naming of the node is convenient for the user but causes issues when used programmatically.

        Returns:
            str: The command friendly name of the node
        """
        return self.internal_name.replace(" ", replace_with)

    @classmethod
    def _get_internal_name_from_command_name(cls, command_name: str) -> str:
        """
        Replace Magnus specific character (%) with whitespace.
        The opposite of _command_friendly_name.

        Args:
            command_name (str): The command friendly node name

        Returns:
            str: The internal name of the step
        """
        return command_name.replace(defaults.COMMAND_FRIENDLY_CHARACTER, " ")

    @classmethod
    def _resolve_map_placeholders(cls, name: str, map_variable: TypeMapVariable = None) -> str:
        """
        If there is no map step used, then we just return the name as we find it.

        If there is a map variable being used, replace every occurrence of the map variable placeholder with
        the value sequentially.

        For example:
        1). dag:
              start_at: step1
              steps:
                step1:
                    type: map
                    iterate_on: y
                    iterate_as: y_i
                    branch:
                        start_at: map_step1
                        steps:
                            map_step1: # internal_name step1.placeholder.map_step1
                                type: task
                                command: a.map_func
                                command_type: python
                                next: map_success
                            map_success:
                                type: success
                            map_failure:
                                type: fail

            and if y is ['a', 'b', 'c'].

            This method would be called 3 times with map_variable = {'y_i': 'a'}, map_variable = {'y_i': 'b'} and
            map_variable = {'y_i': 'c'} corresponding to the three branches.

        For nested map branches, we would get the map_variables ordered hierarchically.

        Args:
            name (str): The name to resolve
            map_variable (dict): The dictionary of map variables

        Returns:
            [str]: The resolved name
        """
        if not map_variable:
            return name

        for _, value in map_variable.items():
            name = name.replace(defaults.MAP_PLACEHOLDER, str(value), 1)

        return name

    def _get_step_log_name(self, map_variable: TypeMapVariable = None) -> str:
        """
        For every step in the dag, there is a corresponding step log name.
        This method returns the step log name in dot path convention.

        All node types except a map state has a "static" defined step_log names and are equivalent to internal_name.
        For nodes belonging to map state, the internal name has a placeholder that is replaced at runtime.

        Args:
            map_variable (dict): If the node is of type map, the names are based on the current iteration state of the
            parameter.

        Returns:
            str: The dot path name of the step log name
        """
        return self._resolve_map_placeholders(self.internal_name, map_variable=map_variable)

    def _get_branch_log_name(self, map_variable: TypeMapVariable = None) -> str:
        """
        For nodes that are internally branches, this method returns the branch log name.
        The branch log name is in dot path convention.

        For nodes that are not map, the internal branch name is equivalent to the branch name.
        For map nodes, the internal branch name has a placeholder that is replaced at runtime.

        Args:
            map_variable (dict): If the node is of type map, the names are based on the current iteration state of the
            parameter.

        Returns:
            str: The dot path name of the branch log
        """
        return self._resolve_map_placeholders(self.internal_branch_name, map_variable=map_variable)

    def __str__(self) -> str:  # pragma: no cover
        """
        String representation of the node.

        Returns:
            str: The string representation of the node.
        """
        return f"Node of type {self.node_type} and name {self.internal_name}"

    @abstractmethod
    def _get_on_failure_node(self) -> str:
        """
        If the node defines a on_failure node in the config, return this or None.

        The naming is relative to the dag, the caller is supposed to resolve it to the correct graph

        Returns:
            str: The on_failure node defined by the dag or ''
        This is a base implementation which the BaseNode does not satisfy
        """
        ...

    @abstractmethod
    def _get_next_node(self) -> str:
        """
        Return the next node as defined by the config.

        Returns:
            str: The node name, relative to the dag, as defined by the config
        """
        ...

    @abstractmethod
    def _is_terminal_node(self) -> bool:
        """
        Returns whether a node has a next node

        Returns:
            bool: True or False of whether there is next node.
        """
        ...

    @abstractmethod
    def _get_catalog_settings(self) -> Dict[str, Any]:
        """
        If the node defines a catalog settings, return it or None

        Returns:
            dict: catalog settings defined as per the node or None
        """
        ...

    @abstractmethod
    def _get_branch_by_name(self, branch_name: str):
        """
        Retrieve a branch by name.

        The name is expected to follow a dot path convention.

        Args:
            branch_name (str): [description]

        Raises:
            Exception: [description]
        """
        ...

    def _get_neighbors(self) -> List[str]:
        """
        Gets the connecting neighbor nodes, either the "next" node or "on_failure" node.

        Returns:
            list: List of connected neighbors for a given node. Empty if terminal node.
        """
        neighbors = []
        try:
            next_node = self._get_next_node()
            neighbors += [next_node]
        except exceptions.TerminalNodeError:
            pass

        try:
            fail_node = self._get_on_failure_node()
            if fail_node:
                neighbors += [fail_node]
        except exceptions.TerminalNodeError:
            pass

        return neighbors

    @abstractmethod
    def _get_executor_config(self, executor_type: str) -> str:
        """
        Return the executor config of the node, if defined, or empty dict

        Args:
            executor_type (str): The executor type that the config refers to.

        Returns:
            dict: The executor config, if defined or an empty dict
        """
        ...

    @abstractmethod
    def _get_max_attempts(self) -> int:
        """
        The number of max attempts as defined by the config or 1.

        Returns:
            int: The number of maximum retries as defined by the config or 1.
        """
        ...

    @abstractmethod
    def execute(self, mock=False, map_variable: TypeMapVariable = None, **kwargs) -> StepAttempt:
        """
        The actual function that does the execution of the command in the config.

        Should only be implemented for task, success, fail and as-is and never for
        composite nodes.

        Args:
            executor (magnus.executor.BaseExecutor): The executor class
            mock (bool, optional): Don't run, just pretend. Defaults to False.
            map_variable (str, optional): The value of the map iteration variable, if part of a map node.
                Defaults to ''.

        Raises:
            NotImplementedError: Base class, hence not implemented.
        """
        ...

    @abstractmethod
    def execute_as_graph(self, map_variable: TypeMapVariable = None, **kwargs):
        """
        This function would be called to set up the execution of the individual
        branches of a composite node.

        Function should only be implemented for composite nodes like dag, map, parallel.

        Args:
            executor (magnus.executor.BaseExecutor): The executor.

        Raises:
            NotImplementedError: Base class, hence not implemented.
        """
        ...

    @abstractmethod
    def fan_out(self, map_variable: TypeMapVariable = None, **kwargs):
        """
        This function would be called to set up the execution of the individual
        branches of a composite node.

        Function should only be implemented for composite nodes like dag, map, parallel.

        Args:
            executor (magnus.executor.BaseExecutor): The executor.
            map_variable (str, optional): The value of the map iteration variable, if part of a map node.

        Raises:
            Exception: If the node is not a composite node.
        """
        ...

    @abstractmethod
    def fan_in(self, map_variable: TypeMapVariable = None, **kwargs):
        """
        This function would be called to tear down the execution of the individual
        branches of a composite node.

        Function should only be implemented for composite nodes like dag, map, parallel.

        Args:
            executor (magnus.executor.BaseExecutor): The executor.
            map_variable (str, optional): The value of the map iteration variable, if part of a map node.

        Raises:
            Exception: If the node is not a composite node.
        """
        ...

    @classmethod
    @abstractmethod
    def parse_from_config(cls, config: Dict[str, Any]) -> "BaseNode":
        """
        Parse the config from the user and create the corresponding node.

        Args:
            config (Dict[str, Any]): The config of the node from the yaml or from the sdk.

        Returns:
            BaseNode: The corresponding node.
        """
        ...


# --8<-- [end:docs]
class TraversalNode(BaseNode):
    next_node: str = Field(serialization_alias="next")
    on_failure: str = Field(default="")
    overrides: Dict[str, str] = Field(default_factory=dict)

    def _get_on_failure_node(self) -> str:
        """
        If the node defines a on_failure node in the config, return this or None.

        The naming is relative to the dag, the caller is supposed to resolve it to the correct graph

        Returns:
            str: The on_failure node defined by the dag or ''
        This is a base implementation which the BaseNode does not satisfy
        """
        return self.on_failure

    def _get_next_node(self) -> str:
        """
        Return the next node as defined by the config.

        Returns:
            str: The node name, relative to the dag, as defined by the config
        """

        return self.next_node

    def _is_terminal_node(self) -> bool:
        """
        Returns whether a node has a next node

        Returns:
            bool: True or False of whether there is next node.
        """
        return False

    def _get_executor_config(self, executor_type) -> str:
        return self.overrides.get(executor_type) or ""


class CatalogStructure(BaseModel):
    model_config = ConfigDict(extra="forbid")  # Need to forbid

    get: List[str] = Field(default_factory=list)
    put: List[str] = Field(default_factory=list)


class ExecutableNode(TraversalNode):
    catalog: Optional[CatalogStructure] = Field(default=None)
    max_attempts: int = Field(default=1, ge=1)

    def _get_catalog_settings(self) -> Dict[str, Any]:
        """
        If the node defines a catalog settings, return it or None

        Returns:
            dict: catalog settings defined as per the node or None
        """
        if self.catalog:
            return self.catalog.model_dump()
        return {}

    def _get_max_attempts(self) -> int:
        return self.max_attempts

    def _get_branch_by_name(self, branch_name: str):
        raise Exception("This is an executable node and does not have branches")

    def execute_as_graph(self, map_variable: TypeMapVariable = None, **kwargs):
        raise Exception("This is an executable node and does not have a graph")

    def fan_in(self, map_variable: TypeMapVariable = None, **kwargs):
        raise Exception("This is an executable node and does not have a fan in")

    def fan_out(self, map_variable: TypeMapVariable = None, **kwargs):
        raise Exception("This is an executable node and does not have a fan out")


class CompositeNode(TraversalNode):
    def _get_catalog_settings(self) -> Dict[str, Any]:
        """
        If the node defines a catalog settings, return it or None

        Returns:
            dict: catalog settings defined as per the node or None
        """
        raise Exception("This is a composite node and does not have a catalog settings")

    def _get_max_attempts(self) -> int:
        raise Exception("This is a composite node and does not have a max_attempts")

    def execute(self, mock=False, map_variable: TypeMapVariable = None, **kwargs) -> StepAttempt:
        raise Exception("This is a composite node and does not have an execute function")


class TerminalNode(BaseNode):
    def _get_on_failure_node(self) -> str:
        raise exceptions.TerminalNodeError()

    def _get_next_node(self) -> str:
        raise exceptions.TerminalNodeError()

    def _is_terminal_node(self) -> bool:
        return True

    def _get_catalog_settings(self) -> Dict[str, Any]:
        raise exceptions.TerminalNodeError()

    def _get_branch_by_name(self, branch_name: str):
        raise exceptions.TerminalNodeError()

    def _get_executor_config(self, executor_type) -> str:
        raise exceptions.TerminalNodeError()

    def _get_max_attempts(self) -> int:
        return 1

    def execute_as_graph(self, map_variable: TypeMapVariable = None, **kwargs):
        raise exceptions.TerminalNodeError()

    def fan_in(self, map_variable: TypeMapVariable = None, **kwargs):
        raise exceptions.TerminalNodeError()

    def fan_out(self, map_variable: TypeMapVariable = None, **kwargs):
        raise exceptions.TerminalNodeError()

    @classmethod
    def parse_from_config(cls, config: Dict[str, Any]) -> "TerminalNode":
        return cls(**config)



# mypy: ignore-errors
# The above should be done until https://github.com/python/mypy/issues/8823
from enum import Enum
from typing import Any, Dict, Mapping, Optional, Union

from typing_extensions import TypeAlias

# TODO: This is not the correct way to do this.
try:  # pragma: no cover
    from typing import TypedDict  # type: ignore[unused-ignore]
except ImportError:  # pragma: no cover
    from typing_extensions import TypedDict  # type: ignore[unused-ignore]


NAME = "magnus"
LOGGER_NAME = "magnus"

# CLI settings
LOG_LEVEL = "WARNING"


class EXECUTION_PLAN(Enum):
    """
    The possible execution plans for a magnus job.
    """

    CHAINED = "chained"  #  121 relationship between run log and the dag.
    UNCHAINED = "unchained"  # Only captures execution of steps, no relation.
    INTERACTIVE = "interactive"  # used for interactive sessions


# Type definitions
class ServiceConfig(TypedDict):
    type: str
    config: Mapping[str, Any]


class MagnusConfig(TypedDict, total=False):
    run_log_store: Optional[ServiceConfig]
    secrets: Optional[ServiceConfig]
    catalog: Optional[ServiceConfig]
    executor: Optional[ServiceConfig]
    experiment_tracker: Optional[ServiceConfig]


TypeMapVariable: TypeAlias = Optional[Dict[str, Union[str, int, float]]]


# Config file environment variable
MAGNUS_CONFIG_FILE = "MAGNUS_CONFIG_FILE"
MAGNUS_RUN_TAG = "MAGNUS_RUN_TAG"

# Interaction settings
TRACK_PREFIX = "MAGNUS_TRACK_"
STEP_INDICATOR = "_STEP_"
PARAMETER_PREFIX = "MAGNUS_PRM_"
MAP_VARIABLE = "MAGNUS_MAP_VARIABLE"
VARIABLE_PREFIX = "MAGNUS_VAR_"
ENV_RUN_ID = "MAGNUS_RUN_ID"
ATTEMPT_NUMBER = "MAGNUS_STEP_ATTEMPT"

# STATUS progression
# For Branch, CREATED -> PROCESSING -> SUCCESS OR FAIL
# For a step, CREATED -> TRIGGERED ->  PROCESSING -> SUCCESS OR FAIL
CREATED = "CREATED"
PROCESSING = "PROCESSING"
SUCCESS = "SUCCESS"
FAIL = "FAIL"
TRIGGERED = "TRIGGERED"

# Node and Command settings
COMMAND_TYPE = "python"
NODE_SPEC_FILE = "node_spec.yaml"
COMMAND_FRIENDLY_CHARACTER = "%"
DEFAULT_CONTAINER_CONTEXT_PATH = "/opt/magnus/"
DEFAULT_CONTAINER_DATA_PATH = "data/"
DEFAULT_CONTAINER_OUTPUT_PARAMETERS = "parameters.json"

# Default services
DEFAULT_EXECUTOR = ServiceConfig(type="local", config={})
DEFAULT_RUN_LOG_STORE = ServiceConfig(type="buffered", config={})
DEFAULT_CATALOG = ServiceConfig(type="file-system", config={})
DEFAULT_SECRETS = ServiceConfig(type="do-nothing", config={})
DEFAULT_EXPERIMENT_TRACKER = ServiceConfig(type="do-nothing", config={})

# Map state
MAP_PLACEHOLDER = "map_variable_placeholder"

# Dag node
DAG_BRANCH_NAME = "dag"

# RUN settings
RANDOM_RUN_ID_LEN = 6
MAX_TIME = 86400  # 1 day in seconds

# User extensions
USER_CONFIG_FILE = "magnus-config.yaml"

# Executor settings
ENABLE_PARALLEL = False

# RUN log store settings
LOG_LOCATION_FOLDER = ".run_log_store"

# Dag node
DAG_BRANCH_NAME = "dag"

# Data catalog settings
CATALOG_LOCATION_FOLDER = ".catalog"
COMPUTE_DATA_FOLDER = "."

# Secrets settings
DOTENV_FILE_LOCATION = ".env"


# Docker settings
DOCKERFILE_NAME = "Dockerfile"
DOCKERFILE_CONTENT = r"""# Python 3.8 Image without Dependecies
FROM python:3.8

LABEL maintainer="mesanthu@gmail.com"

RUN apt-get update && apt-get install -y --no-install-recommends \
    git \
    && rm -rf /var/lib/apt/lists/*

${INSTALL_STYLE}

ENV VIRTUAL_ENV=/opt/venv
RUN python -m virtualenv --python=/usr/local/bin/python $VIRTUAL_ENV
ENV PATH="$VIRTUAL_ENV/bin:$PATH"

${COPY_CONTENT}
WORKDIR /app

${INSTALL_REQUIREMENTS}
"""
GIT_ARCHIVE_NAME = "git_tracked"
LEN_SHA_FOR_TAG = 8


class ENTRYPOINT(Enum):
    """
    The possible container entrypoint types.
    """

    USER = "user"
    SYSTEM = "system"


## Logging settings

LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": True,
    "formatters": {
        "standard": {"format": "%(asctime)s [%(levelname)s] %(name)s: %(message)s"},
        "magnus_formatter": {"format": "%(message)s", "datefmt": "[%X]"},
    },
    "handlers": {
        "default": {
            "formatter": "standard",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stdout",  # Default is stderr
        },
        "magnus_handler": {
            "formatter": "magnus_formatter",
            "class": "rich.logging.RichHandler",
            "rich_tracebacks": True,
        },
    },
    "loggers": {
        "": {
            "handlers": ["default"],
            "propagate": True,
        },  # Root logger
        LOGGER_NAME: {"handlers": ["magnus_handler"], "propagate": False},
    },
}



import logging
from abc import ABC, abstractmethod
from typing import List, Optional

from pydantic import BaseModel, ConfigDict

import magnus.context as context
from magnus import defaults
from magnus.datastore import DataCatalog

logger = logging.getLogger(defaults.LOGGER_NAME)


# --8<-- [start:docs]


class BaseCatalog(ABC, BaseModel):
    """
    Base Catalog class definition.

    All implementations of the catalog handler should inherit and extend this class.
    """

    service_name: str = ""
    service_type: str = "catalog"
    model_config = ConfigDict(extra="forbid")

    @property
    def _context(self):
        return context.run_context

    @property
    def compute_data_folder(self) -> str:
        return defaults.COMPUTE_DATA_FOLDER

    @abstractmethod
    def get(self, name: str, run_id: str, compute_data_folder: str = "", **kwargs) -> List[DataCatalog]:
        """
        Get the catalog item by 'name' for the 'run id' and store it in compute data folder.

        The catalog location should have been created before you can get from it.

        Args:
            name (str): The name of the catalog item
            run_id (str): The run_id of the run.
            compute_data_folder (str, optional): The compute data folder. Defaults to magnus default (data/)

        Raises:
            NotImplementedError: Base class, hence not implemented

        Returns:
            List(object) : A list of catalog objects
        """
        raise NotImplementedError

    @abstractmethod
    def put(
        self,
        name: str,
        run_id: str,
        compute_data_folder: str = "",
        synced_catalogs: Optional[List[DataCatalog]] = None,
        **kwargs,
    ) -> List[DataCatalog]:
        """
        Put the file by 'name' from the 'compute_data_folder' in the catalog for the run_id.

        If previous syncing has happened and the file has not been changed, we do not sync again.

        Args:
            name (str): The name of the catalog item.
            run_id (str): The run_id of the run.
            compute_data_folder (str, optional): The compute data folder. Defaults to magnus default (data/)
            synced_catalogs (dict, optional): Any previously synced catalogs. Defaults to None.

        Raises:
            NotImplementedError: Base class, hence not implemented

        Returns:
            List(object) : A list of catalog objects
        """
        raise NotImplementedError

    @abstractmethod
    def sync_between_runs(self, previous_run_id: str, run_id: str):
        """
        Given run_id of a previous run, sync them to the catalog of the run given by run_id

        Args:
            previous_run_id (str): The run id of the previous run
            run_id (str): The run_id to which the data catalogs should be synced to.

        Raises:
            NotImplementedError: Base class, hence not implemented
        """
        raise NotImplementedError


# --8<-- [end:docs]


class DoNothingCatalog(BaseCatalog):
    """
    A Catalog handler that does nothing.

    Example config:

    catalog:
      type: do-nothing

    """

    service_name: str = "do-nothing"

    def get(self, name: str, run_id: str, compute_data_folder: str = "", **kwargs) -> List[DataCatalog]:
        """
        Does nothing
        """
        logger.info("Using a do-nothing catalog, doing nothing in get")
        return []

    def put(
        self,
        name: str,
        run_id: str,
        compute_data_folder: str = "",
        synced_catalogs: Optional[List[DataCatalog]] = None,
        **kwargs,
    ) -> List[DataCatalog]:
        """
        Does nothing
        """
        logger.info("Using a do-nothing catalog, doing nothing in put")
        return []

    def sync_between_runs(self, previous_run_id: str, run_id: str):
        """
        Does nothing
        """
        logger.info("Using a do-nothing catalog, doing nothing while sync between runs")
        ...



class RunLogExistsError(Exception):  # pragma: no cover
    """
    Exception class
    Args:
        Exception ([type]): [description]
    """

    def __init__(self, run_id):
        super().__init__()
        self.message = f"Run id for {run_id} is already found in the datastore"


class RunLogNotFoundError(Exception):  # pragma: no cover
    """
    Exception class
    Args:
        Exception ([type]): [description]
    """

    def __init__(self, run_id):
        super().__init__()
        self.message = f"Run id for {run_id} is not found in the datastore"


class StepLogNotFoundError(Exception):  # pragma: no cover
    """
    Exception class
    Args:
        Exception ([type]): [description]
    """

    def __init__(self, run_id, name):
        super().__init__()
        self.message = f"Step log for {name} is not found in the datastore for Run id: {run_id}"


class BranchLogNotFoundError(Exception):  # pragma: no cover
    """
    Exception class
    Args:
        Exception ([type]): [description]
    """

    def __init__(self, run_id, name):
        super().__init__()
        self.message = f"Branch log for {name} is not found in the datastore for Run id: {run_id}"


class NodeNotFoundError(Exception):  # pragma: no cover
    """
    Exception class
    Args:
        Exception ([type]): [description]
    """

    def __init__(self, name):
        super().__init__()
        self.message = f"Node of name {name} is not found the graph"


class BranchNotFoundError(Exception):  # pragma: no cover
    """
    Exception class
    Args:
        Exception ([type]): [description]
    """

    def __init__(self, name):
        super().__init__()
        self.message = f"Branch of name {name} is not found the graph"


class TerminalNodeError(Exception):  # pragma: no cover
    def __init__(self):
        super().__init__()
        self.message = "Terminal Nodes do not have next node"


class SecretNotFoundError(Exception):  # pragma: no cover
    """
    Exception class
    Args:
        Exception ([type]): [description]
    """

    def __init__(self, secret_name, secret_setting):
        super().__init__()
        self.message = f"No secret found by name:{secret_name} in {secret_setting}"


class ExecutionFailedError(Exception):  # pragma: no cover
    def __init__(self, run_id: str):
        super().__init__()
        self.message = f"Execution failed for run id: {run_id}"



import logging
import os

from magnus import defaults, exceptions
from magnus.secrets import BaseSecrets

logger = logging.getLogger(defaults.LOGGER_NAME)


class EnvSecretsManager(BaseSecrets):
    """
    A secret manager via environment variables.

    This secret manager returns nothing if the key does not match
    """

    service_name: str = "env-secrets-manager"
    prefix: str = ""
    suffix: str = ""

    def get(self, name: str = "", **kwargs) -> str:
        """
        If a name is provided, we look for that in the environment.
        If a environment variable by that name is not found, we raise an Exception.

        If a name is not provided, we return an empty dictionary.

        Args:
            name (str): The name of the secret to retrieve

        Raises:
            Exception: If the secret by the name is not found.

        Returns:
            [type]: [description]
        """

        try:
            return os.environ[f"{self.prefix}{name}{self.suffix}"]
        except KeyError as _e:
            logger.exception(f"Secret {self.prefix}{name}{self.suffix} not found in environment")
            raise exceptions.SecretNotFoundError(secret_name=name, secret_setting="environment") from _e



import json
import logging
from pathlib import Path

from magnus import defaults, exceptions, utils
from magnus.datastore import BaseRunLogStore, RunLog

logger = logging.getLogger(defaults.LOGGER_NAME)


class FileSystemRunLogstore(BaseRunLogStore):
    """
    In this type of Run Log store, we use a file system to store the JSON run log.

    Every single run is stored as a different file which makes it compatible across other store types.

    When to use:
        When locally testing a pipeline and have the need to compare across runs.
        Its fully featured and perfectly fine if your local environment is where you would do everything.

    Do not use:
        If you need parallelization on local, this run log would not support it.

    Example config:

    run_log:
      type: file-system
      config:
        log_folder: The folder to out the logs. Defaults to .run_log_store

    """

    service_name: str = "file-system"
    log_folder: str = defaults.LOG_LOCATION_FOLDER

    @property
    def log_folder_name(self):
        return self.log_folder

    def write_to_folder(self, run_log: RunLog):
        """
        Write the run log to the folder

        Args:
            run_log (RunLog): The run log to be added to the database
        """
        write_to = self.log_folder_name
        utils.safe_make_dir(write_to)

        write_to_path = Path(write_to)
        run_id = run_log.run_id
        json_file_path = write_to_path / f"{run_id}.json"

        with json_file_path.open("w") as fw:
            json.dump(run_log.model_dump(), fw, ensure_ascii=True, indent=4)  # pylint: disable=no-member

    def get_from_folder(self, run_id: str) -> RunLog:
        """
        Look into the run log folder for the run log for the run id.

        If the run log does not exist, raise an exception. If it does, decode it
        as a RunLog and return it

        Args:
            run_id (str): The requested run id to retrieve the run log store

        Raises:
            FileNotFoundError: If the Run Log has not been found.

        Returns:
            RunLog: The decoded Run log
        """
        write_to = self.log_folder_name

        read_from_path = Path(write_to)
        json_file_path = read_from_path / f"{run_id}.json"

        if not json_file_path.exists():
            raise FileNotFoundError(f"Expected {json_file_path} is not present")

        with json_file_path.open("r") as fr:
            json_str = json.load(fr)
            run_log = RunLog(**json_str)  # pylint: disable=no-member
        return run_log

    def create_run_log(
        self,
        run_id: str,
        dag_hash: str = "",
        use_cached: bool = False,
        tag: str = "",
        original_run_id: str = "",
        status: str = defaults.CREATED,
        **kwargs,
    ) -> RunLog:
        """
        # Creates a Run log
        # Adds it to the db
        """

        try:
            self.get_run_log_by_id(run_id=run_id, full=False)
            raise exceptions.RunLogExistsError(run_id=run_id)
        except exceptions.RunLogNotFoundError:
            pass

        logger.info(f"{self.service_name} Creating a Run Log for : {run_id}")
        run_log = RunLog(
            run_id=run_id,
            dag_hash=dag_hash,
            use_cached=use_cached,
            tag=tag,
            original_run_id=original_run_id,
            status=status,
        )
        self.write_to_folder(run_log)
        return run_log

    def get_run_log_by_id(self, run_id: str, full: bool = False, **kwargs) -> RunLog:
        """
        # Returns the run_log defined by id
        # Raises Exception if not found
        """
        try:
            logger.info(f"{self.service_name} Getting a Run Log for : {run_id}")
            run_log = self.get_from_folder(run_id)
            return run_log
        except FileNotFoundError as e:
            raise exceptions.RunLogNotFoundError(run_id) from e

    def put_run_log(self, run_log: RunLog, **kwargs):
        """
        # Puts the run_log into the database
        """
        logger.info(f"{self.service_name} Putting the run log in the DB: {run_log.run_id}")
        self.write_to_folder(run_log)



import logging

from magnus import defaults
from magnus.defaults import TypeMapVariable
from magnus.extensions.executor import GenericExecutor
from magnus.extensions.nodes import TaskNode
from magnus.nodes import BaseNode

logger = logging.getLogger(defaults.LOGGER_NAME)


class LocalExecutor(GenericExecutor):
    """
    In the mode of local execution, we run everything on the local computer.

    This has some serious implications on the amount of time it would take to complete the run.
    Also ensure that the local compute is good enough for the compute to happen of all the steps.

    Example config:
    execution:
      type: local
      config:
        enable_parallel: True or False to enable parallel.

    """

    service_name: str = "local"

    def trigger_job(self, node: BaseNode, map_variable: TypeMapVariable = None, **kwargs):
        """
        In this mode of execution, we prepare for the node execution and execute the node

        Args:
            node (BaseNode): [description]
            map_variable (str, optional): [description]. Defaults to ''.
        """
        self.prepare_for_node_execution()
        self.execute_node(node=node, map_variable=map_variable, **kwargs)

    def execute_node(self, node: BaseNode, map_variable: TypeMapVariable = None, **kwargs):
        """
        For local execution, we just execute the node.

        Args:
            node (BaseNode): _description_
            map_variable (dict[str, str], optional): _description_. Defaults to None.
        """
        self._execute_node(node=node, map_variable=map_variable, **kwargs)

    def execute_job(self, node: TaskNode):
        """
        Set up the step log and call the execute node

        Args:
            node (BaseNode): _description_
        """

        step_log = self._context.run_log_store.create_step_log(node.name, node._get_step_log_name(map_variable=None))

        self.add_code_identities(node=node, step_log=step_log)

        step_log.step_type = node.node_type
        step_log.status = defaults.PROCESSING
        self._context.run_log_store.add_step_log(step_log, self._context.run_id)
        self.execute_node(node=node)

        # Update the run log status
        step_log = self._context.run_log_store.get_step_log(node._get_step_log_name(), self._context.run_id)
        self._context.run_log_store.update_run_log_status(run_id=self._context.run_id, status=step_log.status)



import logging
import os
import shutil
from pathlib import Path
from typing import List, Optional

from magnus import defaults, utils
from magnus.catalog import BaseCatalog
from magnus.datastore import DataCatalog
from magnus.extensions.catalog import is_catalog_out_of_sync

logger = logging.getLogger(defaults.LOGGER_NAME)


class FileSystemCatalog(BaseCatalog):
    """
    A Catalog handler that uses the local file system for cataloging.

    Note: Do not use this if the steps of the pipeline run on different compute environments.

    Example config:

    catalog:
      type: file-system
      config:
        catalog_location: The location to store the catalog.
        compute_data_folder: The folder to source the data from.

    """

    service_name: str = "file-system"
    catalog_location: str = defaults.CATALOG_LOCATION_FOLDER

    def get_catalog_location(self):
        return self.catalog_location

    def get(self, name: str, run_id: str, compute_data_folder: str = "", **kwargs) -> List[DataCatalog]:
        """
        Get the file by matching glob pattern to the name

        Args:
            name ([str]): A glob matching the file name
            run_id ([str]): The run id

        Raises:
            Exception: If the catalog location does not exist

        Returns:
            List(object) : A list of catalog objects
        """
        logger.info(f"Using the {self.service_name} catalog and trying to get {name} for run_id: {run_id}")

        copy_to = self.compute_data_folder
        if compute_data_folder:
            copy_to = compute_data_folder

        copy_to = Path(copy_to)  # type: ignore

        catalog_location = self.get_catalog_location()
        run_catalog = Path(catalog_location) / run_id / copy_to

        logger.debug(f"Copying objects to {copy_to} from the run catalog location of {run_catalog}")

        if not utils.does_dir_exist(run_catalog):
            msg = (
                f"Expected Catalog to be present at: {run_catalog} but not found.\n"
                "Note: Please make sure that some data was put in the catalog before trying to get from it.\n"
            )
            raise Exception(msg)

        # Iterate through the contents of the run_catalog and copy the files that fit the name pattern
        # We should also return a list of data hashes
        glob_files = run_catalog.glob(name)
        logger.debug(f"Glob identified {glob_files} as matches to from the catalog location: {run_catalog}")

        data_catalogs = []
        run_log_store = self._context.run_log_store
        for file in glob_files:
            if file.is_dir():
                # Need not add a data catalog for the folder
                continue

            if str(file).endswith(".execution.log"):
                continue

            relative_file_path = file.relative_to(run_catalog)

            data_catalog = run_log_store.create_data_catalog(str(relative_file_path))
            data_catalog.catalog_handler_location = catalog_location
            data_catalog.catalog_relative_path = str(relative_file_path)
            data_catalog.data_hash = utils.get_data_hash(str(file))
            data_catalog.stage = "get"
            data_catalogs.append(data_catalog)

            # Make the directory in the data folder if required
            Path(copy_to / relative_file_path.parent).mkdir(parents=True, exist_ok=True)
            shutil.copy(file, copy_to / relative_file_path)

            logger.info(f"Copied {file} from {run_catalog} to {copy_to}")

        if not data_catalogs:
            raise Exception(f"Did not find any files matching {name} in {run_catalog}")

        return data_catalogs

    def put(
        self,
        name: str,
        run_id: str,
        compute_data_folder: str = "",
        synced_catalogs: Optional[List[DataCatalog]] = None,
        **kwargs,
    ) -> List[DataCatalog]:
        """
        Put the files matching the glob pattern into the catalog.

        If previously synced catalogs are provided, and no changes were observed, we do not sync them.

        Args:
            name (str): The glob pattern of the files to catalog
            run_id (str): The run id of the run
            compute_data_folder (str, optional): The compute data folder to sync from. Defaults to settings default.
            synced_catalogs (dict, optional): dictionary of previously synced catalogs. Defaults to None.

        Raises:
            Exception: If the compute data folder does not exist.

        Returns:
            List(object) : A list of catalog objects
        """
        logger.info(f"Using the {self.service_name} catalog and trying to put {name} for run_id: {run_id}")

        copy_from = self.compute_data_folder
        if compute_data_folder:
            copy_from = compute_data_folder
        copy_from = Path(copy_from)  # type: ignore

        catalog_location = self.get_catalog_location()
        run_catalog = Path(catalog_location) / run_id
        utils.safe_make_dir(run_catalog)

        logger.debug(f"Copying objects from {copy_from} to the run catalog location of {run_catalog}")

        if not utils.does_dir_exist(copy_from):
            msg = (
                f"Expected compute data folder to be present at: {compute_data_folder} but not found. \n"
                "Note: Magnus does not create the compute data folder for you. Please ensure that the folder exists.\n"
            )
            raise Exception(msg)

        # Iterate through the contents of copy_from and if the name matches, we move them to the run_catalog
        # We should also return a list of datastore.DataCatalog items

        glob_files = copy_from.glob(name)  # type: ignore
        logger.debug(f"Glob identified {glob_files} as matches to from the compute data folder: {copy_from}")

        data_catalogs = []
        run_log_store = self._context.run_log_store
        for file in glob_files:
            if file.is_dir():
                # Need not add a data catalog for the folder
                continue

            relative_file_path = file.relative_to(".")

            data_catalog = run_log_store.create_data_catalog(str(relative_file_path))
            data_catalog.catalog_handler_location = catalog_location
            data_catalog.catalog_relative_path = run_id + os.sep + str(relative_file_path)
            data_catalog.data_hash = utils.get_data_hash(str(file))
            data_catalog.stage = "put"
            data_catalogs.append(data_catalog)

            if is_catalog_out_of_sync(data_catalog, synced_catalogs):
                logger.info(f"{data_catalog.name} was found to be changed, syncing")

                # Make the directory in the catalog if required
                Path(run_catalog / relative_file_path.parent).mkdir(parents=True, exist_ok=True)
                shutil.copy(file, run_catalog / relative_file_path)
            else:
                logger.info(f"{data_catalog.name} was found to be unchanged, ignoring syncing")

        if not data_catalogs:
            raise Exception(f"Did not find any files matching {name} in {copy_from}")

        return data_catalogs

    def sync_between_runs(self, previous_run_id: str, run_id: str):
        """
        Given the previous run id, sync the catalogs between the current one and previous

        Args:
            previous_run_id (str): The previous run id to sync the catalogs from
            run_id (str): The run_id to which the data catalogs should be synced to.

        Raises:
            Exception: If the previous run log does not exist in the catalog

        """
        logger.info(
            f"Using the {self.service_name} catalog and syncing catalogs"
            "between old: {previous_run_id} to new: {run_id}"
        )

        catalog_location = Path(self.get_catalog_location())
        run_catalog = catalog_location / run_id
        utils.safe_make_dir(run_catalog)

        if not utils.does_dir_exist(catalog_location / previous_run_id):
            msg = (
                f"Catalogs from previous run : {previous_run_id} are not found.\n"
                "Note: Please provision the catalog objects generated by previous run in the same catalog location"
                " as the current run, even if the catalog handler for the previous run was different"
            )
            raise Exception(msg)

        cataloged_files = list((catalog_location / previous_run_id).glob("*"))

        for cataloged_file in cataloged_files:
            if str(cataloged_file).endswith("execution.log"):
                continue
            print(cataloged_file.name)
            if cataloged_file.is_file():
                shutil.copy(cataloged_file, run_catalog / cataloged_file.name)
            else:
                shutil.copytree(cataloged_file, run_catalog / cataloged_file.name)
            logger.info(f"Copied file from: {cataloged_file} to {run_catalog}")



from __future__ import annotations

import json
import logging
import os
from functools import wraps
from typing import Any, ContextManager, Dict, Optional, TypeVar, Union, cast, overload

from pydantic import BaseModel

import magnus.context as context
from magnus import defaults, exceptions, parameters, pickler, utils
from magnus.datastore import RunLog, StepLog

logger = logging.getLogger(defaults.LOGGER_NAME)

CastT = TypeVar("CastT")


def check_context(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        if not context.run_context.executor:
            msg = (
                "There are no active executor and services. This should not have happened and is a bug."
                " Please raise a bug report."
            )
            raise Exception(msg)
        result = func(*args, **kwargs)
        return result

    return wrapper


@check_context
def track_this(step: int = 0, **kwargs):
    """
    Tracks key-value pairs to the experiment tracker.

    The value is dumped as a dict, by alias, if it is a pydantic model.

    Args:
        step (int, optional): The step to track the data at. Defaults to 0.
        **kwargs (dict): The key-value pairs to track.

    Examples:
        >>> track_this(step=0, my_int_param=123, my_float_param=123.45, my_str_param='hello world')
        >>> track_this(step=1, my_int_param=456, my_float_param=456.78, my_str_param='goodbye world')
    """
    prefix = defaults.TRACK_PREFIX

    for key, value in kwargs.items():
        logger.info(f"Tracking {key} with value: {value}")

        if isinstance(value, BaseModel):
            value = value.model_dump(by_alias=True)

        os.environ[prefix + key + f"{defaults.STEP_INDICATOR}{step}"] = json.dumps(value)


@check_context
def set_parameter(**kwargs) -> None:
    """
    Store a set of parameters.

    !!! note
        The parameters are not stored in run log at this point in time.
        They are collected now and stored in the run log after completion of the task.

    Parameters:
        **kwargs (dict): A dictionary of key-value pairs to store as parameters.

    Returns:
        None

    Examples:
        >>> set_parameter(my_int_param=123, my_float_param=123.45, my_bool_param=True, my_str_param='hello world')
        >>> get_parameter('my_int_param', int)
        123
        >>> get_parameter('my_float_param', float)
        123.45
        >>> get_parameter('my_bool_param', bool)
        True
        >>> get_parameter('my_str_param', str)
        'hello world'

        >>> # Example of using Pydantic models
        >>> class MyModel(BaseModel):
        ...     field1: str
        ...     field2: int
        >>> set_parameter(my_model_param=MyModel(field1='value1', field2=2))
        >>> get_parameter('my_model_param', MyModel)
        MyModel(field1='value1', field2=2)

    """
    parameters.set_user_defined_params_as_environment_variables(kwargs)


@overload
def get_parameter(key: str, cast_as: Optional[CastT]) -> CastT:
    ...


@overload
def get_parameter(cast_as: Optional[CastT]) -> CastT:
    ...


@check_context
def get_parameter(key: Optional[str] = None, cast_as: Optional[CastT] = None) -> Union[Dict[str, Any], CastT]:
    """
    Get a parameter by its key.
    If the key is not provided, all parameters will be returned.

    cast_as is not required for JSON supported type (int, float, bool, str).
    For complex nested parameters, cast_as could package them into a pydantic model.
    If cast_as is not provided, the type will remain as dict for nested structures.

    Note that the cast_as pydantic model is the class, not an instance.

    Args:
        key (str, optional): The key of the parameter to retrieve. If not provided, all parameters will be returned.
        cast_as (Type, optional): The type to cast the parameter to. If not provided, the type will remain as it is
            for simple data types (int, float, bool, str). For nested parameters, it would be a dict.

    Raises:
        Exception: If the parameter does not exist and key is not provided.
        ValidationError: If the parameter cannot be cast as pydantic model, when cast_as is provided.

    Examples:
        >>> get_parameter('my_int_param', int)
        123
        >>> get_parameter('my_float_param', float)
        123.45
        >>> get_parameter('my_bool_param', bool)
        True
        >>> get_parameter('my_str_param', str)
        'hello world'
        >>> get_parameter('my_model_param', MyModel)
        MyModel(field1='value1', field2=2)
        >>> get_parameter(cast_as=MyModel)
        MyModel(field1='value1', field2=2)

    """
    params = parameters.get_user_set_parameters(remove=False)

    if not key:
        # Return all parameters
        return cast(CastT, parameters.cast_parameters_as_type(params, cast_as))  # type: ignore

    if key not in params:
        raise Exception(f"Parameter {key} is not set before")

    # Return the parameter value, casted as asked.
    return cast(CastT, parameters.cast_parameters_as_type(params[key], cast_as))  # type: ignore


@check_context
def get_secret(secret_name: str) -> str:
    """
    Retrieve a secret from the secret store.

    Args:
        secret_name (str): The name of the secret to retrieve.

    Raises:
        SecretNotFoundError: If the secret does not exist in the store.

    Returns:
        str: The secret value.
    """
    secrets_handler = context.run_context.secrets_handler
    try:
        return secrets_handler.get(name=secret_name)
    except exceptions.SecretNotFoundError:
        logger.exception(f"No secret by the name {secret_name} found in the store")
        raise


@check_context
def get_from_catalog(name: str, destination_folder: str = ""):
    """
    Get data from the catalog.

    The name can be a wildcard pattern following globing rules.

    Args:
        name (str): The name of the data catalog entry.
        destination_folder (str, optional): The destination folder to download the data to.
            If not provided, the default destination folder set in the catalog will be used.
    """
    if not destination_folder:
        destination_folder = context.run_context.catalog_handler.compute_data_folder

    data_catalog = context.run_context.catalog_handler.get(
        name,
        run_id=context.run_context.run_id,
    )

    if context.run_context.executor._context_step_log:
        context.run_context.executor._context_step_log.add_data_catalogs(data_catalog)
    else:
        logger.warning("Step log context was not found during interaction! The step log will miss the record")


@check_context
def put_in_catalog(filepath: str):
    """
    Add a file or folder to the data catalog.
    You can use wild cards following globing rules.

    Args:
        filepath (str): The path to the file or folder added to the catalog
    """

    data_catalog = context.run_context.catalog_handler.put(
        filepath,
        run_id=context.run_context.run_id,
    )
    if not data_catalog:
        logger.warning(f"No catalog was done by the {filepath}")

    if context.run_context.executor._context_step_log:
        context.run_context.executor._context_step_log.add_data_catalogs(data_catalog)
    else:
        logger.warning("Step log context was not found during interaction! The step log will miss the record")


@check_context
def put_object(data: Any, name: str):
    """
    Serialize and store a python object in the data catalog.

    This function behaves the same as `put_in_catalog`
    but with python objects.

    Args:
        data (Any): The python data object to store.
        name (str): The name to store it against.
    """
    native_pickler = pickler.NativePickler()

    native_pickler.dump(data=data, path=name)
    put_in_catalog(f"{name}{native_pickler.extension}")

    # Remove the file
    os.remove(f"{name}{native_pickler.extension}")


@check_context
def get_object(name: str) -> Any:
    """
    Retrieve and deserialize a python object from the data catalog.

    This function behaves the same as `get_from_catalog` but with
    python objects.

    Returns:
        Any : The object
    """
    native_pickler = pickler.NativePickler()

    get_from_catalog(name=f"{name}{native_pickler.extension}", destination_folder=".")

    try:
        data = native_pickler.load(name)

        # Remove the file
        os.remove(f"{name}{native_pickler.extension}")
        return data
    except FileNotFoundError as e:
        msg = f"No object by the name {name} has been put in the catalog before."
        logger.exception(msg)
        raise e


@check_context
def get_run_id() -> str:
    """
    Returns the run_id of the current run.

    You can also access this from the environment variable `MAGNUS_RUN_ID`.
    """
    return context.run_context.run_id


@check_context
def get_run_log() -> RunLog:
    """
    Returns the run_log of the current run.

    The return is a deep copy of the run log to prevent any modification.
    """
    return context.run_context.run_log_store.get_run_log_by_id(
        context.run_context.run_id,
        full=True,
    ).copy(deep=True)


@check_context
def get_tag() -> str:
    """
    Returns the tag from the environment.

    Returns:
        str: The tag if provided for the run, otherwise None
    """
    return context.run_context.tag


@check_context
def get_experiment_tracker_context() -> ContextManager:
    """
    Return a context session of the experiment tracker.

    You can start to use the context with the python ```with``` statement.
    """
    experiment_tracker = context.run_context.experiment_tracker
    return experiment_tracker.client_context


def start_interactive_session(run_id: str = "", config_file: str = "", tag: str = "", parameters_file: str = ""):
    """
    During interactive python coding, either via notebooks or ipython, you can start a magnus session by calling
    this function. The executor would always be local executor as its interactive.

    If this was called during a pipeline/function/notebook execution, it will be ignored.

    Args:
        run_id (str, optional): The run id to use. Defaults to "" and would be created if not provided.
        config_file (str, optional): The configuration file to use. Defaults to "" and magnus defaults.
        tag (str, optional): The tag to attach to the run. Defaults to "".
        parameters_file (str, optional): The parameters file to use. Defaults to "".
    """

    from magnus import entrypoints, graph  # pylint: disable=import-outside-toplevel

    if context.run_context.executor:
        logger.warn("This is not an interactive session or a session has already been activated.")
        return

    run_id = utils.generate_run_id(run_id=run_id)
    context.run_context = entrypoints.prepare_configurations(
        configuration_file=config_file,
        run_id=run_id,
        tag=tag,
        parameters_file=parameters_file,
        force_local_executor=True,
    )

    executor = context.run_context.executor

    utils.set_magnus_environment_variables(run_id=run_id, configuration_file=config_file, tag=tag)

    context.run_context.execution_plan = defaults.EXECUTION_PLAN.INTERACTIVE.value
    executor.prepare_for_graph_execution()
    step_config = {
        "command": "interactive",
        "command_type": "python",
        "type": "task",
        "next": "success",
    }

    node = graph.create_node(name="interactive", step_config=step_config)
    step_log = context.run_context.run_log_store.create_step_log("interactive", node._get_step_log_name())
    executor.add_code_identities(node=node, step_log=step_log)

    step_log.step_type = node.node_type
    step_log.status = defaults.PROCESSING
    executor._context_step_log = step_log


def end_interactive_session():
    """
    Ends an interactive session.

    Does nothing if the executor is not interactive.
    """

    if not context.run_context.executor:
        logger.warn("There is no active session in play, doing nothing!")
        return

    if context.run_context.execution_plan != defaults.EXECUTION_PLAN.INTERACTIVE.value:
        logger.warn("There is not an interactive session, doing nothing!")
        return

    tracked_data = utils.get_tracked_data()
    set_parameters = parameters.get_user_set_parameters(remove=True)

    step_log = cast(StepLog, context.run_context.executor._context_step_log)
    step_log.user_defined_metrics = tracked_data
    context.run_context.run_log_store.add_step_log(step_log, context.run_context.run_id)

    context.run_context.run_log_store.set_parameters(context.run_context.run_id, set_parameters)

    context.run_context.executor._context_step_log = None
    context.run_context.execution_plan = ""
    context.run_context.executor = None  # type: ignore



import contextlib
import json
import logging
import os
from abc import ABC, abstractmethod
from collections import defaultdict
from typing import Any, ContextManager, Dict, Tuple, Union

from pydantic import BaseModel, ConfigDict

import magnus.context as context
from magnus import defaults
from magnus.utils import remove_prefix

logger = logging.getLogger(defaults.LOGGER_NAME)


def retrieve_step_details(key: str) -> Tuple[str, int]:
    key = remove_prefix(key, defaults.TRACK_PREFIX)
    data = key.split(defaults.STEP_INDICATOR)

    key = data[0].lower()
    step = 0

    if len(data) > 1:
        step = int(data[1])

    return key, step


def get_tracked_data() -> Dict[str, Any]:
    tracked_data: Dict[str, Any] = defaultdict(dict)
    for env_var, value in os.environ.items():
        if env_var.startswith(defaults.TRACK_PREFIX):
            key, step = retrieve_step_details(env_var)

            # print(value, type(value))
            try:
                value = json.loads(value)
            except json.decoder.JSONDecodeError:
                logger.warning(f"Tracker {key} could not be JSON decoded, adding the literal value")

            tracked_data[key][step] = value
            del os.environ[env_var]

    for key, value in tracked_data.items():
        if len(value) == 1:
            tracked_data[key] = value[0]

    return tracked_data


# --8<-- [start:docs]


class BaseExperimentTracker(ABC, BaseModel):
    """
    Base Experiment tracker class definition.
    """

    service_name: str = ""
    service_type: str = "experiment_tracker"

    @property
    def _context(self):
        return context.run_context

    model_config = ConfigDict(extra="forbid")

    @property
    def client_context(self) -> ContextManager:
        """
        Returns the client context.
        """
        return contextlib.nullcontext()

    def publish_data(self, tracked_data: Dict[str, Any]):
        for key, value in tracked_data.items():
            if isinstance(value, dict):
                for key2, value2 in value.items():
                    self.log_metric(key, value2, step=key2)
                continue
            self.log_metric(key, value)

    @abstractmethod
    def log_metric(self, key: str, value: Union[int, float], step: int = 0):
        """
        Sets the metric in the experiment tracking.

        Args:
            key (str): The key against you want to store the value
            value (float): The value of the metric
            step (int): Optional step at which it was recorded

        Raises:
            NotImplementedError: Base class, hence not implemented
        """
        raise NotImplementedError

    @abstractmethod
    def log_parameter(self, key: str, value: Any):
        """
        Logs a parameter in the experiment tracking.

        Args:
            key (str): The key against you want to store the value
            value (any): The value of the metric

        Raises:
            NotImplementedError: Base class, hence not implemented
        """
        pass


# --8<-- [end:docs]


class DoNothingTracker(BaseExperimentTracker):
    """
    A Do nothing tracker
    """

    service_name: str = "do-nothing"

    def log_metric(self, key: str, value: Union[int, float], step: int = 0):
        """
        Sets the metric in the experiment tracking.

        Args:
            key (str): The key against you want to store the value
            value (float): The value of the metric
        """
        ...

    def log_parameter(self, key: str, value: Any):
        """
        Since this is a Do nothing tracker, we don't need to log anything.
        """
        ...
